from adverb_semantics import AdverbSemantics
from shape import  Const, Shape, Tuple, Closure, ConstSlice, AnyScalar
from shape import Slice, Scalar, Add, Sub, Div, Mult
from shape import const, any_scalar
from shape import is_zero, is_one, make_shape

class ShapeSemantics(AdverbSemantics):
  def size_along_axis(self, value, axis):
    assert isinstance(value, Shape)
    return value.dims[axis]

  def is_tuple(self, x):
    return isinstance(x, Tuple)

  def is_none(self, x):
    return isinstance(x, Const) and x.value is None

  def rank(self, value):
    if isinstance(value, Shape):
      return value.rank
    else:
      return 0

  def int(self, x):
    return const(x)

  def bool(self, x):
    return const(x)

  def add(self, x, y):
    if is_zero(x):
      return y
    elif is_zero(y):
      return x
    elif isinstance(x, Const) and isinstance(y, Const):
      return const(x.value + y.value)
    elif isinstance(x, AnyScalar) or isinstance(y, AnyScalar):
      return any_scalar
    else:
      return Add(x,y)

  def sub(self, x, y):
    if is_zero(y):
      return x
    elif isinstance(x, Const) and isinstance(y, Const):
      return const(x.value - y.value)
    elif isinstance(x, AnyScalar) or isinstance(y, AnyScalar):
      return any_scalar
    else:
      return Sub(x, y)

  def mul(self, x, y):
    if is_zero(x) or is_zero(y):
      return const(0)
    elif is_one(x):
      return y
    elif is_one(y):
      return x
    else:
      return Mult(x,y)
    
  def div(self, x, y):
    assert not is_zero(y)
    if is_one(y):
      return x
    elif isinstance(x, Const) and isinstance(y, Const):
      return const(int(x.value / y.value))
    elif isinstance(x, AnyScalar) or isinstance(y, AnyScalar):
      return any_scalar
    else:
      return Div(x, y)

  def shape(self, x):
    if isinstance(x, Shape):
      return Tuple(x.dims)
    else:
      return Tuple(())

  def elt_type(self, x):
    return "DON'T CARE ABOUT ELT TYPES"

  def alloc_array(self, _, dims):
    return make_shape(dims)

  def index(self, arr, idx):
    if isinstance(arr, Scalar):
      return arr
    assert arr.__class__ is Shape
    if isinstance(idx, (Scalar, Slice)):
      indices = [idx]
    elif idx.__class__ is Tuple:
      indices = idx.elts
    else:
      assert False, "Unexpected index: %s" % (idx,)
    result_dims = []
    for (i, curr_idx) in enumerate(indices):
      old_dim = arr.dims[i]
      if curr_idx is None or \
         (isinstance(curr_idx, Const) and curr_idx.value is None):
        result_dims.append(old_dim)
      elif isinstance(curr_idx, Scalar):
        pass
      elif curr_idx.__class__ is ConstSlice:
        result_dims.append(curr_idx.nelts)
      elif curr_idx.__class__ is Shape:
        if len(curr_idx.dims) == 0:
          # same as unknown scalar 
          pass 
        else:
          assert len(curr_idx.dims) == 1, "Indexing by a multi-dimensional array not yet supported"
          result_dims.append(curr_idx.dims[0])
      else:
        assert curr_idx.__class__ is Slice, "Unsupported index %s" % curr_idx

        if curr_idx.start is None:
          lower = const(0)
        elif isinstance(curr_idx.start, Const):
          if curr_idx.start.value is None:
            lower = const(0)
          elif curr_idx.start.value < 0:
            lower = self.sub(old_dim, curr_idx.start)
          else:
            lower = curr_idx.start
        else:
          lower = any_scalar
         
        if curr_idx.stop is None:
          upper = old_dim 
        elif isinstance(curr_idx.stop, Const):
          if curr_idx.stop.value is None:
            upper = old_dim
          elif curr_idx.stop.value < 0:
            upper = self.sub(old_dim, curr_idx.stop)
          else:
            upper = curr_idx.stop
        else:
          upper = any_scalar

        n = self.sub(upper, lower)
        step = curr_idx.step
        if step and \
            isinstance(step, Const) and \
            step.value is not None and \
            step.value != 1:
          n = self.div(n, step)
        result_dims.append(n)
    n_original = len(arr.dims)
    n_idx= len(indices)
    if n_original > n_idx:
      result_dims.extend(arr.dims[n_idx:])

    return make_shape(result_dims)

  def tuple(self, elts):
    return Tuple(tuple(elts))

  def concat_tuples(self, t1, t2):
    return Tuple(t1.elts + t2.elts)

  def setidx(self, arr, idx, v):
    pass

  def loop(self, start_idx, stop_idx, body):
    body(start_idx)

  class Accumulator(object):
    def __init__(self, v):
      self.v = v

    def update(self, new_v):
      self.v = new_v

    def get(self):
      return self.v

  def accumulate_loop(self, start_idx, stop_idx, body, init):
    acc = self.Accumulator(init)
    body(acc, start_idx)
    return acc.get()

  def check_equal_sizes(self, sizes):
    pass

  def slice_value(self, start, stop, step):
    
    # if all elements of the slice are constant 
    # then we can ignore the exact start/stop/step 
    # and track only the number of elements in the 
    # slice
    if start.__class__ is Const and \
       stop.__class__ is Const and \
       stop.value is not None and \
       step.__class__ is Const:
      start_val = start.value
      if start_val is None:
        start_val = 0
      step_val = step.value
      if step_val is None:
        step_val = 1
      nelts = (stop.value - start_val) / step_val
      return ConstSlice(nelts)
    else:
      return Slice(start, stop, step)

  def invoke(self, fn, args):
    if fn.__class__ is Closure:
      args = tuple(fn.args) + tuple(args)
      fn = fn.fn
    import shape_inference
    return shape_inference.symbolic_call(fn, args)

  none = None
  null_slice = slice(None, None, None)

  def identity_function(self, x):
    return x
  
  def ravel(self, x):
    assert isinstance(x, Shape)
    nelts = 1 
    for d in x.dims:
      nelts *= d
    return Shape((nelts,)) 



import os

import llvm.core as llc 

from llvm.core import Builder, ATTR_NO_CAPTURE, Module 
from llvm.core import Type as lltype


import config
import escape_analysis
import llvm_context
import llvm_convert
import llvm_prims
import llvm_types
import prims
import syntax_helpers

from core_types import BoolT, FloatT, SignedT, UnsignedT, ScalarT, NoneT
from core_types import Int32, Int64, PtrT
from llvm_helpers import const, int32, zero 
from llvm_types import llvm_value_type, llvm_ref_type
from syntax import Var, Struct, Index, TypedFn, Attribute

_escape_analysis_cache = {}
class Compiler(object):
  def __init__(self, fundef, llvm_cxt = llvm_context.global_context):
    self.parakeet_fundef = fundef
    if config.opt_stack_allocation:
      self.may_escape = escape_analysis.may_escape(fundef)
    else:
      self.may_escape = None
    self.llvm_context = llvm_cxt
    self.vars = {}
    self.initialized = set([])
    # Initializes the variables dictionary and returns a builder object
    llvm_input_types = map(llvm_ref_type, fundef.input_types)
    llvm_output_type = llvm_ref_type(fundef.return_type)
    llvm_fn_t = lltype.function(llvm_output_type, llvm_input_types)

    self.llvm_fn = self.llvm_context.module.add_function(llvm_fn_t, fundef.name)

    for arg in self.llvm_fn.args:
      if not llvm_types.is_scalar(arg.type):
        arg.add_attribute(ATTR_NO_CAPTURE)

    self.llvm_fn.does_not_throw = True

    self.entry_block, self.entry_builder = self.new_block("entry")
    self._init_vars(self.parakeet_fundef, self.entry_builder)

  def new_block(self, name):
    bb = self.llvm_fn.append_basic_block(name)
    builder = Builder.new(bb)
    return bb, builder

  def _init_vars(self, fundef, builder):
    """
    Create a mapping from variable names to stack locations, these will later be
    converted to SSA variables by the mem2reg pass.
    """
    n_expected = len(fundef.arg_names)
    n_compiled = len(self.llvm_fn.args)
    assert n_compiled == n_expected, \
        "Expected %d args (%s) but compiled code had %d args (%s)" % \
        (n_expected, fundef.arg_names, n_compiled, self.llvm_fn.args)

    for (name, t) in fundef.type_env.iteritems():
      if not name.startswith("$"):
        llvm_t = llvm_ref_type(t)
        stack_val = builder.alloca(llvm_t, name)
        self.vars[name] = stack_val

    for llvm_arg, name in zip(self.llvm_fn.args, fundef.arg_names):
      self.initialized.add(name)
      llvm_arg.name = name
      if name in self.vars:
        builder.store(llvm_arg, self.vars[name])

  def attribute_lookup(self, struct, name, builder):
    """
    Helper for getting the address of an attribute lookup, used both when
    setting and getting attributes
    """
    llvm_struct = self.compile_expr(struct, builder)
    struct_t = struct.type
    field_pos = struct_t.field_pos(name)
    field_type = struct_t.field_type(name)
    indices = [int32(0), int32(field_pos)]
    ptr_name = "%s_ptr" % name
    ptr = builder.gep(llvm_struct, indices, ptr_name)
    return ptr, field_type

  def compile_Var(self, expr, builder):
    name = expr.name
    assert name in self.initialized, "%s uninitialized" % name
    ref = self.vars[expr.name]
    val = builder.load(ref, expr.name + "_val")
    return val

  def compile_Const(self, expr, builder):
    t = expr.type

    if isinstance(t, NoneT):
      return const(0, Int64)
    else:
      assert isinstance(expr.type, ScalarT), \
          "Expected scalar constant but got %s" % expr.type
    return const(expr.value, expr.type)

  def compile_Cast(self, expr, builder):
    llvm_value = self.compile_expr(expr.value, builder)
    return llvm_convert.convert(llvm_value, expr.value.type, expr.type, builder)

  def compile_Struct(self, expr, builder, local = False):
    struct_t = expr.type
    llvm_struct_t = llvm_value_type(struct_t)
    name = expr.type.node_type()
    if local:
      struct_ptr = builder.alloca(llvm_struct_t, name + "_local_ptr")
    else:
      struct_ptr = builder.malloc(llvm_struct_t, name + "_ptr")

    for (i, elt) in enumerate(expr.args):
      field_name, field_type = struct_t._fields_[i]
      assert elt.type == field_type, \
          "Mismatch between expected type %s and given %s for field '%s' " % \
          (field_type, elt.type, field_name)
      elt_ptr = builder.gep(struct_ptr, [int32(0), int32(i)], "field%d_ptr" % i)
      llvm_elt = self.compile_expr(elt, builder)
      builder.store(llvm_elt, elt_ptr)

    return struct_ptr

  def compile_Alloc(self, expr, builder):
    elt_t = expr.elt_type
    llvm_elt_t = llvm_types.llvm_value_type(elt_t)
    n_elts = self.compile_expr(expr.count, builder)
    return builder.malloc_array(llvm_elt_t, n_elts, "data_ptr")

  def compile_Index(self, expr, builder):
    llvm_arr = self.compile_expr(expr.value, builder)
    llvm_index = self.compile_expr(expr.index, builder)
    pointer = builder.gep(llvm_arr, [llvm_index], "elt_pointer")
    elt = builder.load(pointer, "elt", align = 16) #, invariant = True)

    return elt

  def compile_Attribute(self, expr, builder):
    field_ptr, _ = \
        self.attribute_lookup(expr.value, expr.name, builder)
    field_value = builder.load(field_ptr, "%s_value" % expr.name)
    return field_value

  def compile_TypedFn(self, expr, builder):
    (target_fn, _, _) = compile_fn(expr)
    return target_fn

  def compile_Call(self, expr, builder):
    assert expr.fn.__class__ is TypedFn
    typed_fundef = expr.fn
    (target_fn, _, _) = compile_fn(typed_fundef)
    arg_types = syntax_helpers.get_types(expr.args)
    llvm_args = [self.compile_expr(arg, builder) for arg in expr.args]
    assert len(arg_types) == len(llvm_args)
    return builder.call(target_fn, llvm_args, 'call_result')

  def cmp(self, prim, t, llvm_x, llvm_y, builder, result_name = None):
    if result_name is None:
      result_name = prim.name + "_result"

    if isinstance(t, FloatT):
      cmp_op = llvm_prims.float_comparisons[prim]
      return builder.fcmp(cmp_op, llvm_x, llvm_y, result_name)
    elif isinstance(t, SignedT):
      cmp_op = llvm_prims.signed_int_comparisons[prim]
      return builder.icmp(cmp_op, llvm_x, llvm_y, result_name)
    else:
      assert isinstance(t, (BoolT, UnsignedT)), "Unexpected type for comparison %s: %s" % (prim, t)
      cmp_op = llvm_prims.unsigned_int_comparisons[prim]
      return builder.icmp(cmp_op, llvm_x, llvm_y, result_name)

  
  def lt(self, t, llvm_x, llvm_y, builder, result_name = None ):
    return self.cmp(prims.less, t, llvm_x, llvm_y, builder, result_name)
  
  def lte(self, t, llvm_x, llvm_y, builder, result_name = None ):
    return self.cmp(prims.less_equal, t, llvm_x, llvm_y, builder, result_name)
  
  def neq(self, t, llvm_x, llvm_y, builder, result_name = None):
    return self.cmp(prims.not_equal, t, llvm_x, llvm_y, builder, result_name)
  
  def sub(self, t, x, y, builder, result_name = "sub"):
    if isinstance(t, FloatT):
      return builder.fsub(x,y,result_name)
    else:
      return builder.sub(x,y,result_name)
    
  def add(self, t, x, y, builder, result_name = "add"):
    if isinstance(t, FloatT):
      return builder.fadd(x,y,result_name)
    else:
      return builder.add(x,y,result_name)
    
  def neg(self, x, builder):
    if isinstance(x.type, llc.IntegerType):
      return builder.neg(x, "neg")
    else:
      return builder.fsub(zero(x.type), x, "neg") 
  def prim(self, prim, t, llvm_args, builder, result_name = None):
    if result_name is None:
      result_name = prim.name + "_result"

    if isinstance(prim, prims.Cmp):
      bit = self.cmp(prim, t, llvm_args[0], llvm_args[1], builder)
      return llvm_convert.to_bool(bit,builder)
    
    elif prim == prims.maximum:
      x, y = llvm_args
      bit = self.cmp(prims.greater_equal, t, x, y, builder)
      return builder.select(bit, x, y)
    
    elif prim == prims.minimum:
      x,y = llvm_args
      bit = self.cmp(prims.less_equal, t, x, y, builder)
      return builder.select(bit, x, y)
    
    elif prim == prims.negative:
      return self.neg(llvm_args[0], builder)
    
    # python's remainder is weird in that it preserve's the sign of 
    # the second argument, whereas LLVM's srem/frem operators preserve
    # the sign of the first 
    elif prim == prims.mod:
      x,y = llvm_args 
      if isinstance(t, (UnsignedT, BoolT)):
        return builder.urem(llvm_args[0], llvm_args[1], "modulo")
      elif isinstance(t, SignedT): 
        rem = builder.srem(x,y, "modulo")
      else:
        assert isinstance(t, FloatT)
        rem = builder.frem(llvm_args[0], llvm_args[1], "modulo")

      y_is_negative = self.lt(t, y, zero(y.type), builder, "first_arg_negative")
      rem_is_negative = self.lt(t, rem, zero(rem.type), builder, "rem_is_negative")
      y_nonzero = self.neq(t, y, zero(y.type), builder, "first_arg_nonzero")
      rem_nonzero = self.neq(t,x,zero(x.type),builder,"rem_nonzero")
      neither_zero = builder.and_(y_nonzero, rem_nonzero, "neither_zero")
      diff_signs = builder.xor(y_is_negative, rem_is_negative, "different_signs")
      should_flip = builder.and_(neither_zero, diff_signs, "should_flip") 
      flipped_rem = self.add(t, y, rem, builder, "flipped_rem") 
      return builder.select(should_flip, flipped_rem, rem)
        
    elif isinstance(prim, prims.Arith) or isinstance(prim, prims.Bitwise):
      if isinstance(t, FloatT):
        instr = llvm_prims.float_binops[prim]
      elif isinstance(t, SignedT):
        instr = llvm_prims.signed_binops[prim]
      elif isinstance(t, UnsignedT):
        instr = llvm_prims.unsigned_binops[prim]
      else:
        assert isinstance(t, BoolT)
        instr = llvm_prims.bool_binops[prim]
      op = getattr(builder, instr)
      return op(name = result_name, *llvm_args)

    elif isinstance(prim, prims.Logical):
      if prim == prims.logical_and:
        return builder.and_(name = result_name, *llvm_args)
      elif prim == prims.logical_not:
        return builder.not_(name = result_name, *llvm_args)
      else:
        assert prim == prims.logical_or
        return builder.or_(name = result_name, *llvm_args)
      
    elif prim == prims.abs:
      x = llvm_args[0]
      bit = self.cmp(prims.greater_equal, t,  x, zero(x.type), builder, "gt_zero")
      neg_value = self.neg(x, builder)
      return builder.select(bit, x, neg_value)
    elif isinstance(prim, prims.Float): 
      llvm_op = llvm_prims.get_float_unary_op(prim, t)
      return builder.call(llvm_op, llvm_args)
    
    else:
      assert False, "UNSUPPORTED PRIMITIVE: %s" % prim

  def compile_PrimCall(self, expr, builder):
    args = expr.args
    # type specialization should have made types of arguments uniform,
    # so we only need to check the type of the first arg
    t = args[0].type
    llvm_args = [self.compile_expr(arg, builder) for arg in args]
    return self.prim(expr.prim, t, llvm_args, builder)

  def compile_expr(self, expr, builder):
    method_name = "compile_" + expr.node_type()
    return getattr(self, method_name)(expr, builder)

  def compile_Assign(self, stmt, builder):
    rhs_t = stmt.rhs.type
    # special case for locally allocated structs
    if self.may_escape is not None and \
       stmt.lhs.__class__ is Var and \
       stmt.rhs.__class__ is Struct and \
       stmt.lhs.name  not in self.may_escape:
      value = self.compile_Struct(stmt.rhs, builder, local = True)
    else:
      value = self.compile_expr(stmt.rhs, builder)
    if stmt.lhs.__class__ is Var:
      name = stmt.lhs.name
      lhs_t = stmt.lhs.type
      self.initialized.add(name)
      ref = self.vars[name]
    elif stmt.lhs.__class__ is Index:
      ptr_t = stmt.lhs.value.type
      assert isinstance(ptr_t, PtrT), \
          "Expected pointer, got %s" % ptr_t
      lhs_t = ptr_t.elt_type
      base_ptr = self.compile_expr(stmt.lhs.value, builder)
      index = self.compile_expr(stmt.lhs.index, builder)
      index = llvm_convert.from_signed(index, Int32, builder)
      ref = builder.gep(base_ptr, [index], "elt_ptr")
    else:
      assert stmt.lhs.__class__ is Attribute, \
          "Unexpected LHS: %s" % stmt.lhs
      struct = stmt.lhs.value
      ref, lhs_t = self.attribute_lookup(struct, stmt.lhs.name, builder)

    assert lhs_t == rhs_t, \
        "Type mismatch between LHS %s and RHS %s" % (lhs_t, rhs_t)

    builder.store(value, ref)
    return builder, False

  def compile_ExprStmt(self, stmt, builder):
    self.compile_expr(stmt.value, builder)
    return builder, False

  def compile_Return(self, stmt, builder):
    ret_val = self.compile_expr(stmt.value, builder)
    builder.ret(ret_val)
    return builder, True

  def compile_merge_left(self, phi_nodes, builder):
    for name, (left, _) in phi_nodes.iteritems():
      ref = self.vars[name]
      self.initialized.add(name)
      value = self.compile_expr(left, builder)
      builder.store(value, ref)

  def compile_merge_right(self, phi_nodes, builder):
    for name, (_, right) in phi_nodes.iteritems():
      ref = self.vars[name]
      self.initialized.add(name)
      value = self.compile_expr(right, builder)
      builder.store(value, ref)

  def compile_ForLoop(self, stmt, builder):
    # first compile the starting, boundary, and
    # increment values for the loop counter
    start = self.compile_expr(stmt.start, builder)
    stop = self.compile_expr(stmt.stop, builder)
    step = self.compile_expr(stmt.step, builder)

    # get the memory slot associated with the loop counter
    loop_var = self.vars[stmt.var.name]

    builder.store(start, loop_var)
    self.initialized.add(stmt.var.name)

    # ...and we'll need its Parakeet type later on
    # for calls to 'cmp' and 'prim'
    loop_var_t = stmt.var.type

    # any phi-bound variables should be initialized to their
    # starting values
    self.compile_merge_left(stmt.merge, builder)

    loop_bb, body_start_builder = self.new_block("loop_body")
    after_bb, after_builder = self.new_block("after_loop")

    # WARNING: Assuming loop is always increasing,
    # only enter the loop if we're less than the stopping value
    enter_cond = self.cmp(prims.less, loop_var_t,  start, stop,
                          builder, "enter_cond")
    builder.cbranch(enter_cond, loop_bb, after_bb)

    # TODO: what should we do if the body always ends in a return statement?
    body_end_builder, body_always_returns = \
      self.compile_block(stmt.body, body_start_builder)

    counter_at_end = body_end_builder.load(loop_var)
    # increment the loop counter
    incr = self.prim(prims.add, loop_var_t,
                     [counter_at_end, step],
                     body_end_builder,  "incr_loop_var")
    body_end_builder.store(incr, loop_var)
    self.compile_merge_right(stmt.merge, body_end_builder)

    exit_cond = self.cmp(prims.less, loop_var_t, incr, stop,
                         body_end_builder, "exit_cond")
    body_end_builder.cbranch(exit_cond, loop_bb, after_bb)
    # WARNING: what if the loop doesn't run? Should
    # we still be returning 'body_always_returns'?
    return after_builder, body_always_returns

  def compile_While(self, stmt, builder):
    # current flow ----> loop --------> exit--> after
    #    |                       skip------------|
    #    |----------------------/

    self.compile_merge_left(stmt.merge, builder)
    loop_bb, body_start_builder = self.new_block("loop_body")

    after_bb, after_builder = self.new_block("after_loop")
    enter_cond = self.compile_expr(stmt.cond, builder)
    enter_cond = llvm_convert.to_bit(enter_cond, builder)
    builder.cbranch(enter_cond, loop_bb, after_bb)

    body_end_builder, body_always_returns = \
        self.compile_block(stmt.body, body_start_builder)
    if not body_always_returns:
      exit_bb, exit_builder = self.new_block("loop_exit")
      self.compile_merge_right(stmt.merge, body_end_builder)
      repeat_cond = self.compile_expr(stmt.cond, body_end_builder)
      repeat_cond = llvm_convert.to_bit(repeat_cond, body_end_builder)
      body_end_builder.cbranch(repeat_cond, loop_bb, exit_bb)
      exit_builder.branch(after_bb)

    return after_builder, False

  def compile_If(self, stmt, builder):
    cond = self.compile_expr(stmt.cond, builder)
    cond = llvm_convert.to_bit(cond, builder)

    if len(stmt.true) == 0 and len(stmt.false) == 0:
      # if nothing happens in the loop bodies, just
      # emit select instructions
      for (name, (true_expr, false_expr)) in stmt.merge.iteritems():
        ref = self.vars[name]
        self.initialized.add(name)
        true_val = self.compile_expr(true_expr, builder)
        false_val = self.compile_expr(false_expr, builder)
        select_val = builder.select(cond, true_val, false_val)
        builder.store(select_val, ref)
      return builder, False
    else:
      # compile the two possible branches as distinct basic blocks
      # and then wire together the control flow with branches
      true_bb, true_builder = self.new_block("if_true")
      after_true, true_always_returns = \
        self.compile_block(stmt.true, true_builder)

      false_bb, false_builder = self.new_block("if_false")
      after_false, false_always_returns = \
          self.compile_block(stmt.false, false_builder)

      builder.cbranch(cond, true_bb, false_bb)

      # compile phi nodes as assignments and then branch
      # to the continuation block
      self.compile_merge_left(stmt.merge, after_true)
      self.compile_merge_right(stmt.merge, after_false)

      # if both branches return then there is no point
      # making a new block for more code
      # did both branches end in a return?
      both_always_return = true_always_returns and false_always_returns
      if both_always_return:
        return None, True
      after_bb, after_builder = self.new_block("if_after")
      if not true_always_returns:
        after_true.branch(after_bb)
      if not false_always_returns:
        after_false.branch(after_bb)
      return after_builder, False

  def compile_Comment(self, stmt, builder):
    return builder, False

  def compile_stmt(self, stmt, builder):
    """
    Translate an SSA statement into LLVM. Every translation function returns a
    builder pointing to the end of the current basic block and a boolean
    indicating whether every branch of control flow in that statement ends in a
    return. The latter is needed to avoid creating empty basic blocks, which
    were causing some mysterious crashes inside LLVM.
    """
    method_name = "compile_" + stmt.node_type()
    return getattr(self, method_name)(stmt, builder)

  def compile_block(self, stmts, builder):
    for stmt in stmts:
      builder, always_returns = self.compile_stmt(stmt, builder)
      if always_returns:
        return builder, always_returns
    return builder, False

  def compile_body(self, body):
    return self.compile_block(body, builder = self.entry_builder)

compiled_functions = {}
def compile_fn(fundef):
  key = fundef.name, fundef.copied_by
  if key in compiled_functions:
    return compiled_functions[key]
  
  if config.print_lowered_function:
    print
    print "=== Lowered function ==="
    print
    print repr(fundef)
    print

  compiler = Compiler(fundef)
  compiler.compile_body(fundef.body)
  if config.print_unoptimized_llvm:
    print "=== LLVM before optimizations =="
    print
    print compiler.llvm_context.module
    print
  compiler.llvm_context.run_passes(compiler.llvm_fn)

  if config.print_optimized_llvm:
    print "=== LLVM after optimizations =="
    print
    print compiler.llvm_context.module
    print

  if config.print_x86:
    print "=== Generated assembly =="
    print
    start_printing = False
    w,r = os.popen2("llc")
    w.write(str(compiler.llvm_context.module))
    w.close()
    assembly_str = r.read()
    r.close()
    for l in assembly_str.split('\n'):
      if l.strip().startswith('.globl'):
        if start_printing:
          break
        elif fundef.name in l:
          start_printing = True
      if start_printing:
        print l

  result = (compiler.llvm_fn, fundef, compiler.llvm_context.exec_engine)
  compiled_functions[key] = result
  return result



import syntax_helpers

from array_type import ArrayT
from closure_type import ClosureT
from core_types import ScalarT
from shape import Closure, Scalar, Var
from traversal import Traversal
from tuple_type import TupleT
from shape import AnyScalar, any_scalar, computable_dim

class ArgConverter(Traversal):
  def __init__(self, codegen):
    self.codegen = codegen
    self.var_counter = 0
    self.env = {}

  def fresh_var(self):
    n = self.var_counter
    self.var_counter += 1
    return Var(n)

  def bind(self, scalar_value):
    v = self.fresh_var()
    self.env[v] = scalar_value

  def convert(self, x):
    t = x.type
    if isinstance(t, ScalarT):
      self.bind(x)
    elif isinstance(t, ArrayT):
      shape = self.codegen.shape(x)
      shape_elts = self.codegen.tuple_elts(shape)
      return self.convert_list(shape_elts)
    elif isinstance(t, TupleT):
      elts = self.codegen.tuple_elts(x)
      self.convert_list(elts)
    elif isinstance(t, ClosureT):
      closure_elts = self.codegen.closure_elts(x)
      return Closure(t.fn, self.convert_list(closure_elts))
    else:
      assert False, "[shape_codegen] Not supported %s : %s" % (x,x.type)

  def convert_list(self, xs):
    for x in xs:
      self.convert(x)

class ShapeCodegen(Traversal):
  def __init__(self, codegen, exprs):
    self.codegen = codegen
    conv = ArgConverter(codegen)
    self.exprs = exprs
    conv.convert_list(exprs)
    self.env = conv.env

  def visit_Var(self, v):
    return self.env[v]

  def visit_Const(self, v):
    return syntax_helpers.const(v.value)

  def visit_Shape(self, v):
    assert len(v.dims) > 0, "Encountered empty shape"
    assert all(computable_dim(d) for d in v.dims), \
        "Symbolic shape '%s' has non-constant dimensions" % (v,)
    return self.codegen.tuple([self.visit(d) for d in v.dims])

  def visit_Dim(self, v):
    return self.codegen.tuple_proj(self.visit(v.array), v.dim)

  def visit_AnyScalar(self, v):    
    assert False, "Can't generate shape expression for unknown scalar"

  def visit_Tuple(self, v):
    return self.codegen.tuple(self.visit(e) for e in v.elts)

  def binop(self, op_name, v):
    if v.x.__class__ is AnyScalar or v.y.__class__ is AnyScalar:
      return any_scalar 
    x = self.visit(v.x)
    y = self.visit(v.y)
    op = getattr(self.codegen, op_name)
   
    return op(x,y)

  def visit_Sub(self, v):
    return self.binop('sub', v)

  def visit_Add(self, v):
    return self.binop('add', v)

  def visit_Mult(self, v):
    return self.binop('mult', v)

  def visit_Div(self, v):
    return self.binop('div', v)

  def visit_Mod(self, v):
    return self.binop('mod', v)

  def visit_Closure(self, v):
    assert False, "Unexpected closure in result shape: %s" % (v,)

def make_shape_expr(codegen, symbolic_shape, input_exprs):
  """
  Given a codegen object we're currently using to create a function, and a
  symbolic result shape of a function call (along with the input expressions
  that went into the function) generate a code expression for the shape of the
  result
  """

  if isinstance(symbolic_shape, Scalar):
    return codegen.tuple([])
  shape_codegen = ShapeCodegen(codegen, input_exprs)
  return shape_codegen.visit(symbolic_shape)



import llvm.core as llcore
import llvm_types

from core_types import FloatT, SignedT, UnsignedT, BoolT, IntT
from llvm_helpers import zero, one
from llvm_types import int1_t, int8_t, llvm_value_type, nbytes

def to_bit(llvm_value, builder):
  llvm_t = llvm_value.type

  if llvm_t == int1_t:
    return llvm_value
  if isinstance(llvm_t, llcore.IntegerType):
    return builder.icmp(llcore.ICMP_NE, llvm_value, zero(llvm_t), "ne_zero")
  else:
    return builder.fcmp(llcore.FCMP_ONE, llvm_value, zero(llvm_t), "ne_zero")

def from_bit(llvm_value, new_ptype, builder):
  llvm_t = llvm_value_type(new_ptype)
  name = "%s.cast.%s" % (llvm_value.name, new_ptype)
  return builder.select(llvm_value, one(llvm_t), zero(llvm_t), name)

def to_bool(llvm_value, builder):
  """
  bools are stored as bytes, if you need to use a boolean value for control flow
  convert it to a bit instead
  """

  bit = to_bit(llvm_value, builder)
  return builder.zext(bit, int8_t, "bool_val")

def from_float(llvm_value, new_ptype, builder):
  """Convert from an LLVM float value to some other LLVM scalar type"""

  dest_llvm_type = llvm_value_type(new_ptype)
  dest_name = "%s.cast_%s" % (llvm_value.name, new_ptype)

  if isinstance(new_ptype, FloatT):
    if llvm_types.nbytes(llvm_value.type) <= new_ptype.nbytes:
      return builder.fpext(llvm_value, dest_llvm_type, dest_name)
    else:
      return builder.fptrunc(llvm_value, dest_llvm_type, dest_name)
  elif isinstance(new_ptype, SignedT):
    return builder.fptosi(llvm_value, dest_llvm_type, dest_name)
  elif isinstance(new_ptype, UnsignedT):
    return builder.fptoui(llvm_value, dest_llvm_type, dest_name)
  else:
    return to_bool(llvm_value, builder)

def from_signed(llvm_value, new_ptype, builder):
  """Convert from an LLVM float value to some other LLVM scalar type"""

  dest_llvm_type = llvm_value_type(new_ptype)
  dest_name = "%s.cast_%s" % (llvm_value.name, new_ptype)

  if isinstance(new_ptype, FloatT):
    return builder.sitofp(llvm_value, dest_llvm_type, dest_name)
  elif isinstance(new_ptype, BoolT):
    return to_bool(llvm_value, builder)
  else:
    assert isinstance(new_ptype, IntT)
    nbits = llvm_value.type.width
    nbytes = nbits / 8

    if nbytes == new_ptype.nbytes:
      return builder.bitcast(llvm_value, dest_llvm_type, dest_name)
    elif nbytes < new_ptype.nbytes:
      return builder.zext(llvm_value, dest_llvm_type, dest_name)
    else:
      return builder.trunc(llvm_value, dest_llvm_type, dest_name)

def from_unsigned(llvm_value, new_ptype, builder):
  """Convert from an LLVM float value to some other LLVM scalar type"""

  dest_llvm_type = llvm_value_type(new_ptype)
  dest_name = "%s.cast_%s" % (llvm_value.name, new_ptype)

  if isinstance(new_ptype, FloatT):
    return builder.uitofp(llvm_value, dest_llvm_type, dest_name)
  elif isinstance(new_ptype, BoolT):
    return to_bool(llvm_value, builder)
  else:
    assert isinstance(new_ptype, IntT)
    nbytes = llvm_value.type.width / 8

    if nbytes == new_ptype.nbytes:
      return builder.bitcast(llvm_value, dest_llvm_type, dest_name)
    elif nbytes < new_ptype.nbytes:
      return builder.zext(llvm_value, dest_llvm_type, dest_name)
    else:
      return builder.trunc(llvm_value, dest_llvm_type, dest_name)

def convert(llvm_value, old_ptype, new_ptype, builder):
  """
  Given an LLVM value and two parakeet types, generate the instruction to
  perform the conversion
  """

  if old_ptype == new_ptype:
    return llvm_value

  if isinstance(old_ptype, FloatT):
    return from_float(llvm_value, new_ptype, builder)
  elif isinstance(old_ptype, SignedT):
    return from_signed(llvm_value, new_ptype, builder)
  else:
    assert isinstance(old_ptype, (BoolT, UnsignedT))
    return from_unsigned(llvm_value, new_ptype, builder)



import syntax

from closure_type import ClosureT
from collect_vars import collect_var_names, collect_binding_names
from escape_analysis import may_alias
from scoped_set import ScopedSet
from syntax import Var, Assign, Return, While, If, Index
from syntax_visitor import SyntaxVisitor
from transform import Transform
from tuple_type import TupleT

class Find_LICM_Candidates(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.mutable_types = None
    self.volatile_vars = ScopedSet()
    self.depends_on = {}
    self.safe_to_move = set([])
    self.curr_block_id = None
    self.block_contains_return = set([])
    self.may_alias = None

  def visit_fn(self, fn):
    self.volatile_vars.push(fn.arg_names)
    self.may_alias = may_alias(fn)
    SyntaxVisitor.visit_fn(self, fn)
    return self.safe_to_move

  def mark_safe_assignments(self, block, volatile_set):
    for stmt in block:
      klass = stmt.__class__
      if klass is Assign and \
         stmt.lhs.__class__ is Var:
        name = stmt.lhs.name
        dependencies = self.depends_on.get(name, set([]))
        volatile = name in volatile_set or \
                   any(d in volatile_set for d in dependencies)
        if not volatile:
          self.safe_to_move.add(name)
      # just in case there are Returns in nested control flow
      # we should probably avoid changing the performance characteristics
      # by pulling out statements which will never run
      elif klass is If:
        if id(stmt.true) in self.block_contains_return or \
           id(stmt.false) in self.block_contains_return:
          break
      elif klass is While:
        if id(stmt.body) in self.block_contains_return:
          break
      elif klass is Return:
        break

  def mark_curr_block_returns(self):
    self.block_contains_return.add(self.curr_block_id)

  def does_block_return(self, block):
    return id(block) in self.block_contains_return

  def visit_Return(self, stmt):
    self.mark_curr_block_returns()

  def visit_block(self, stmts):
    self.curr_block_id = id(stmts)
    SyntaxVisitor.visit_block(self, stmts)

  def visit_merge(self, merge, both_branches = True):
    pass

  def visit_ForLoop(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    self.volatile_vars.add(stmt.var.name)
    SyntaxVisitor.visit_ForLoop(self, stmt)
    if self.does_block_return(stmt.body):
      self.block_contains_return()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.body, volatile_in_scope)

  def visit_While(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    SyntaxVisitor.visit_While(self, stmt)
    if self.does_block_return(stmt.body):
      self.block_contains_return()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.body, volatile_in_scope)

  def visit_Var(self, expr):
    self.volatile_vars.add(expr.name)

  def visit_If(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    self.visit_expr(stmt.cond)
    SyntaxVisitor.visit_If(self, stmt)
    if self.does_block_return(stmt.true) or self.does_block_return(stmt.false):
      self.mark_curr_block_returns()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.true, volatile_in_scope)
    self.mark_safe_assignments(stmt.false, volatile_in_scope)

  def is_mutable_alloc(self, expr):
    c = expr.__class__
    safe_types = (TupleT, ClosureT)
    return c is syntax.Alloc or \
           c is syntax.Array or \
           c is syntax.ArrayView or \
           c is syntax.Slice or  \
           (c is syntax.Struct and not isinstance(expr.type, safe_types))

  def visit_Assign(self, stmt):
     
    lhs_names = collect_binding_names(stmt.lhs)
    rhs_names = collect_var_names(stmt.rhs)
   
    for x in lhs_names:
      dependencies = self.depends_on.get(x, set([]))
      dependencies.update(rhs_names)
      self.depends_on[x] = dependencies

    if any(x in self.volatile_vars for x in rhs_names):
      self.volatile_vars.update(lhs_names)
    elif self.is_mutable_alloc(stmt.rhs):
      if len(lhs_names) == 1 and \
         len(self.may_alias.get(lhs_names[0], [])) <= 1:
        pass
      else:
        self.volatile_vars.update(lhs_names)
    # mark any array writes as volatile 
    if stmt.lhs.__class__ is Index:
      assert stmt.lhs.value.__class__ is Var
      self.volatile_vars.add(stmt.lhs.value.name)
      #print 
      #print "STMT", stmt
      #print "lhs names", lhs_names 
      #print "rhs names", rhs_names 
      #print "volatile vars", self.volatile_vars
      #print 
      

class LoopInvariantCodeMotion(Transform):
  def __init__(self):
    Transform.__init__(self)

  def pre_apply(self, fn):
    self.analysis = Find_LICM_Candidates()
    self.safe_to_move = self.analysis.visit_fn(fn)
    self.binding_depth = {}
    self.mark_binding_depths(fn.arg_names, 0)

  def mark_binding_depths(self, names, depth_offset = 0):
    curr_depth = len(self.blocks._blocks) - 1 + depth_offset
    for name in names:
      self.binding_depth[name] = curr_depth

  def transform_ForLoop(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    self.mark_binding_depths([stmt.var.name], 1)
    return Transform.transform_ForLoop(self, stmt)

  def transform_While(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    return Transform.transform_While(self, stmt)

  def transform_If(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    return Transform.transform_If(self, stmt)

  def transform_Assign(self, stmt):
    # TODO: Allow possibility of indexing into variables that are read-only
    if stmt.lhs.__class__ is Var and stmt.rhs.__class__ is not Index:
      name = stmt.lhs.name
      if name in self.safe_to_move:
        deps = self.analysis.depends_on[name]

        if all(d in self.binding_depth for d in deps):
          if len(deps) > 0:
            target_level = max(self.binding_depth[d] for d in deps)
          else:
            target_level = 0
 
          if target_level >= 0 and target_level < self.blocks.depth():
            self.blocks._blocks[target_level].append(stmt)
            self.binding_depth[name] = target_level
            return None
    self.mark_binding_depths(collect_binding_names(stmt.lhs))
    return stmt



import array_type
import core_types

from collect_vars import collect_binding_names
from core_types import NoneT, NoneType
from syntax import Expr, Tuple, Var, Index, Closure 
from syntax_helpers import get_types
from syntax_visitor import SyntaxVisitor

class Verify(SyntaxVisitor):
  def __init__(self, fn):
    SyntaxVisitor.__init__(self)
    self.fn = fn
    self.bound = set(fn.arg_names)

  def bind_var(self, name):
    assert name not in self.bound, \
        "Error: variable %s already bound" % name
    self.bound.add(name)
    assert name in self.fn.type_env, \
        "Assigned variable %s has no entry in type dictionary" % name

  def phi_value_ok(self, lhs_name, v):
    assert isinstance(v, Expr), \
       "Invalid expression in phi node: %s" % (v,)
    self.visit_expr(v)
    assert v.type is not None, \
       "Error in phi node: value %s has no type annotation" % v
    t = self.fn.type_env[lhs_name]
    assert v.type == t, \
       "Invalid type annotation on %v, expected %s but got %s" % (v,t,v.type)

  def visit_merge_loop_start(self, phi_nodes):
    for (k, (v,_)) in phi_nodes.iteritems():
      self.bind_var(k)
      self.phi_value_ok(k,v)

  def visit_merge_loop_repeat(self, phi_nodes):
    for (k, (left_value, right_value)) in phi_nodes.iteritems():
      self.phi_value_ok(k, left_value)
      self.phi_value_ok(k, right_value)

  def visit_merge_if(self, phi_nodes):
    for (k, (left_value, right_value)) in phi_nodes.iteritems():
      self.bind_var(k)
      self.phi_value_ok(k, left_value)
      self.phi_value_ok(k, right_value)

  def visit_Var(self, expr):
    assert expr.name in self.bound, \
        "Variable %s used before assignment" % expr.name
    assert expr.name in self.fn.type_env, \
        "Variable %s has no entry in the type dictionary" % expr.name
    assert expr.type == self.fn.type_env[expr.name], \
        "Variable %s in fn %s should have type %s but annotated with %s" % \
        (expr.name, self.fn, self.fn.type_env[expr.name], expr.type)

  def visit_PrimCall(self, expr):
    self.visit_expr_list(expr.args)

    for arg in expr.args:
      assert arg.type is not None, \
          "Expected type annotation for %s" % (arg, )
      assert isinstance(arg.type, core_types.ScalarT), \
          "Can't call primitive %s with argument %s of non-scalar type %s" % \
          (expr.prim, arg, arg.type)

  def visit_expr(self, expr):
    assert expr is not None
    assert expr.type is not None, \
      "Missing type annotation on %s" % expr 
    SyntaxVisitor.visit_expr(self, expr)

  def visit_ExprStmt(self, stmt):
    self.visit_expr(stmt.value)
    assert stmt.value.type and stmt.value.type.__class__ is NoneT, \
      "Expected effectful expression %s to have type %s but instead got %s" % \
      (stmt.value, NoneType, stmt.value.type)
  
  def check_fn_args(self, fn, args, arg_types = None):
    if arg_types is None: 
      arg_types = get_types(args)
    n_given = len(arg_types)
    n_expected = len(fn.input_types)
    assert n_given == n_expected, \
        "Arity mismatch when calling %s, expected %d but got %d" % \
        (fn.name, n_expected, n_given)
    for (i, arg_name) in enumerate(fn.arg_names):
      given_t = arg_types[i]
      expected_t = fn.type_env[arg_name]
      signature_t = fn.input_types[i] 
      assert expected_t == signature_t, \
          "Function %s has inconsistent types %s and %s for arg %s" % \
          (fn.name, expected_t, signature_t, arg_name)
      assert given_t == expected_t, \
          "Given argument '%s' : %s doesn't matched expected '%s' : %s in %s" % \
          (args[i], given_t, arg_name, expected_t, fn.name)  
  
  
  def visit_Call(self, expr):
    if expr.fn.__class__ is Closure: 
      closure_elts = tuple(expr.fn.args) 
      fn = expr.fn.fn 
    else:
      fn = expr.fn
      closure_elts = ()
    try:
      self.check_fn_args(fn, closure_elts + tuple(expr.args))
    except:
      print "[verify] Errors in function call %s" %expr 
      raise 
    
  def visit_Map(self, expr):
    if expr.fn.__class__ is Closure: 
      closure_elts = tuple(expr.fn.args) 
      fn = expr.fn.fn 
    else:
      fn = expr.fn
      closure_elts = ()
    elt_types = [array_type.lower_rank(arg.type, 1) for arg in expr.args]
    arg_types = tuple(get_types(closure_elts)) + tuple(elt_types)
    args = tuple(closure_elts) + tuple(expr.args)                      
    self.check_fn_args(fn, args, arg_types)
  
  def visit_Return(self, stmt):
    self.visit_expr(stmt.value)
    #print "RET TYPE", stmt.value.type, stmt.value.type == True, stmt.value.type == False
    assert stmt.value.type is not None and stmt.value.type == self.fn.return_type, \
        "Incorrect return type in %s: ret value %s, expected %s but got %s" % \
        (self.fn.name, stmt.value, self.fn.return_type, stmt.value.type)

  def visit_lhs(self, lhs):
    c = lhs.__class__
    assert c in (Var, Tuple, Index), \
       "Invalid LHS of assignment"
    if c is Tuple:
      for elt in lhs.elts:
        self.visit_lhs(elt)

  def visit_Assign(self, stmt):
    assert stmt.lhs.type is not None, \
        "Missing LHS type for assignment %s" % stmt
    assert stmt.rhs.type is not None, \
        "Missing RHS type for assignment %s" % stmt
    self.visit_lhs(stmt.lhs)
    lhs_names = collect_binding_names(stmt.lhs)
    for lhs_name in lhs_names:
      self.bind_var(lhs_name)
    self.visit_expr(stmt.rhs)

    assert stmt.lhs.type == stmt.rhs.type, \
        "Mismatch between LHS type %s and RHS %s in '%s'" % \
        (stmt.lhs.type, stmt.rhs.type, stmt)

  def visit_ForLoop(self, stmt):
    assert stmt.var.__class__ is Var
    self.bind_var(stmt.var.name)
    self.visit_expr(stmt.var)

    assert stmt.start.type == stmt.var.type
    self.visit_expr(stmt.start)
    self.visit_merge_loop_start(stmt.merge)

    assert stmt.stop.type == stmt.var.type
    self.visit_expr(stmt.stop)
    self.visit_block(stmt.body)
    
    assert stmt.step.type == stmt.var.type
    self.visit_expr(stmt.step)
    self.visit_merge_loop_repeat(stmt.merge)

  def visit_stmt(self, stmt):
    assert stmt is not None
    SyntaxVisitor.visit_stmt(self, stmt)

  def visit_TypedFn(self, fn):
    return verify(fn)

def verify(fn):
  n_input_types = len(fn.input_types)
  n_arg_names = len(fn.arg_names)
  assert n_input_types == n_arg_names, \
     "Function %s has %d input types and %d input names" % \
     (fn.name, n_input_types, n_arg_names)
  for arg_name, input_t in zip(fn.arg_names, fn.input_types):
    assert arg_name in fn.type_env, \
        "Input %s has no entry in %s's type environment" % \
        (arg_name, fn.name)
    t = fn.type_env[arg_name]
    assert input_t == t, \
        "Mismatch between %s's input type %s and the arg %s's type %s" % \
        (fn.name, input_t, arg_name, t)
  try:    
    Verify(fn).visit_block(fn.body)
  except:
    print "[verify] Errors in body of function", repr(fn)
    raise 


import numpy as np 
from timer import compare_perf

# Simple convolution of 3x3 patches from a given array x
# by a 3x3 array of filter weights
 
def conv_3x3_trim(x, weights):
  return np.array([[(x[i-1:i+2, j-1:j+2]*weights).sum() 
                    for j in xrange(1, x.shape[1] -2)]
                    for i in xrange(1, x.shape[0] -2)])
 

x = np.random.randn(500,500)
w = np.random.randn(3,3)

compare_perf(conv_3x3_trim, [x,w])


# Simple convolution of 3x3 patches from a given array x
# by a 3x3 array of filter weights
 
def conv_3x3_trim_loops(x, weights):
  result = np.zeros_like(x)
  for i in xrange(1,x.shape[0]-2):
    for j in xrange(1,x.shape[1]-2):
      window = x[i-1:i+2, j-1:j+2]
      result[i, j] = (window*weights).sum()
  return result


compare_perf(conv_3x3_trim_loops, [x,w])



from traversal import Traversal 
import types 
import numpy as np 

def transform_value(x):
  """
  Replace arrays with their shapes, 
  and recursively replace any instances of arrays
  in data structures like tuples also with their shapes
  """
  if isinstance(x, np.ndarray):
    return x.shape 
  elif isinstance(x, list):
    return np.array(x).shape 
  elif isinstance(x, tuple):
    return tuple(transform_value(elt) for elt in x)

  elif isinstance(x, (int, long, float, complex, types.NoneType)):
    return x
  else:
    return None

class EvalShape(Traversal):

  def __init__(self, input_values):
    
    self.inputs = []
    for x in input_values:
      y = transform_value(x)
      if hasattr(y, '__iter__'):
        self.inputs.extend(y)
      else:
        self.inputs.append(y)
  
  def visit_Var(self, v):
    return self.inputs[v.num]  
  
  def visit_Const(self, v):
    return v.value 

  def visit_AnyScalar(self, v):
    return None 
    
  def visit_Shape(self, v):
    dims = self.visit_tuple(v.dims)
    assert all(isinstance(d, int) for d in dims)
    return dims
    
  def visit_Dim(self, v):
    return self.visit(v.array)[v.dim]
    
  def visit_Tuple(self, v):
    return self.visit_tuple(v.elts)
    
  def visit_Sub(self, v):
    return self.visit(v.x) - self.visit(v.y)
  
  def visit_Add(self, v):
    return self.visit(v.x) + self.visit(v.y)
  
  def visit_Mult(self, v):
    return self.visit(v.x) * self.visit(v.y)  
  
  def visit_Div(self, v):
    return self.visit(v.x) / self.visit(v.y)  
  
  def visit_Mod(self, v):
    return self.visit(v.x) % self.visit(v.y)  
  
  def visit_Closure(self, v):
    return v.fn, self.visit_tuple(v.args)

def eval_shape(symbolic_shape, input_values):
  evaluator = EvalShape(input_values)
  result = evaluator.visit(symbolic_shape)
  if not isinstance(result, tuple):
    return () 
  else:
    assert all(isinstance(elt, int) for elt in result)
    return result
   

def result_shape(typed_fn, input_values):
  import shape_inference
  symbolic_shape = shape_inference.call_shape_expr(typed_fn)
  return eval_shape(symbolic_shape, input_values)

      
  


import numpy as np

def sqr_dists(X,Y):
  return np.array([[np.sum( (x-y) ** 2) for x in X] for y in Y])

def sqr_dists_loops(X,Y):
  m,n = X.shape[0], Y.shape[0]
  D = np.zeros((m,n), dtype=X.dtype)
  for i in xrange(m):
    for j in xrange(n):
      D[i,j] = np.sum( (X[i, :] -Y[j, :]) ** 2)
  return D

ndims = 10
nsamples = 10**4
nclusters = 50
X = np.random.randn(nsamples, ndims)
Y = np.random.randn(nclusters, ndims)




from timer import timer 

print 
print "Computing distances between %d and %d %s vectors of length %d" % \
  (nsamples, nclusters, X.dtype, ndims)

# 
# Parakeet 
# 

import parakeet

parakeet_dists = parakeet.jit(sqr_dists)

with timer('Parakeet (comprehensions) #1'):
  parakeet_dists(X,Y)

with timer('Parakeet (comprehensions) #2'):
  parakeet_dists(X,Y)

parakeet_dists_loops = parakeet.jit(sqr_dists_loops)

with timer('Parakeet (loops) #1'):
  parakeet_dists_loops(X,Y)

with timer('Parakeet (loops) #2'):
  parakeet_dists_loops(X,Y)



#
# Pure Python 
# 
from timer import timer

with timer('Python (comprehensions)'):
  sqr_dists(X,Y)

with timer('Python (loops)'):
  sqr_dists_loops(X,Y)



#
# Numba 
#

import numba

#
# Numba's @autojit just like Parakeet's @jit
#
numba_dists = numba.autojit(sqr_dists)

with timer('Numba (comprehensions) #1'):
  numba_dists(X,Y)

with timer('Numba (comprehensions) #2'):
  numba_dists(X,Y)

numba_dists_loops = numba.autojit(sqr_dists_loops)

with timer('Numba (loops) #1'):
  numba_dists_loops(X,Y)

with timer('Numba (loops) #2'):
  numba_dists_loops(X,Y)




import array_type
import closure_type
import core_types
import names
import prims
import shape_codegen
import shape_inference
import syntax
import syntax_helpers

from array_type import ArrayT, SliceT
from core_types import ScalarT, Int32, Int64, NoneT, Type, StructT
from closure_type import ClosureT, make_closure_type
from nested_blocks import NestedBlocks
from syntax import AllocArray, ForLoop, Comment 
from syntax import Var, Assign, Closure, Attribute, PrimCall
from syntax import Index, Const, TypedFn, Struct, ClosureElt, Cast
from syntax import TupleProj, Tuple, Alloc, Slice, While, Fn
from syntax import ArrayView
from syntax_helpers import get_types, wrap_constants, wrap_if_constant, \
                           one_i64, zero, zero_i64, \
                           one, const_int, const_bool
from tuple_type import TupleT, make_tuple_type

class Codegen(object):
  def __init__(self):
    self.type_env = {}
    self.blocks = NestedBlocks()

    # cut down the number of created nodes by
    # remembering which tuple variables we've created
    # and looking up their elements directly
    self.tuple_elt_cache = {}

  def comment(self, text):
    self.blocks.append(Comment(text))
    
  def fresh_var(self, t, prefix = "temp"):
    assert prefix is not None
    assert t is not None, "Type required for new variable %s" % prefix
    ssa_id = names.fresh(prefix)
    self.type_env[ssa_id] = t
    return Var(ssa_id, type = t)

  def fresh_i32(self, prefix = "temp"):
    return self.fresh_var(Int32, prefix)

  def fresh_i64(self, prefix = "temp"):
    return self.fresh_var(Int64, prefix)

  def insert_stmt(self, stmt):
    self.blocks.append_to_current(stmt)

  def assign(self, lhs, rhs):
    self.insert_stmt(Assign(lhs, rhs))

  def temp_name(self, expr):
    c = expr.__class__
    if c is PrimCall:
      return expr.prim.name
    elif c is Attribute:
      if expr.value.__class__ is Var:
        return names.original(expr.value.name) + "_" + expr.name
      else:
        return expr.name
    elif c is Attribute:
      if expr.value.__class__ is Var:
        return "%s_%s" % (expr.value.name, expr.name)
      else:
        return expr.name
    elif c is Index:
      idx_t = expr.index.type
      if isinstance(idx_t, SliceT) or \
         (isinstance(idx_t, TupleT) and \
          any(isinstance(elt_t, SliceT) for elt_t in idx_t.elt_types)):
        return "slice"
      else:
        return "elt"
    elif c is TupleProj:
      if expr.tuple.__class__ is Var:
        original = names.original(expr.tuple.name)
        return "%s%d_elt%d" % (original, names.versions[original], expr.index)
      else:
        return "tuple_elt%d" % expr.index
    else:
      return "temp"

  def is_simple(self, expr):
    c = expr.__class__
    return c is Var or \
           c is Const or \
           (c is Tuple and len(expr.elts) == 0) or \
           (c is Struct and len(expr.args) == 0) or \
           (c is Closure and len(expr.args) == 0)



  def assign_temp(self, expr, name = None):
    if self.is_simple(expr):
      return expr
    if name is None:
      name = self.temp_name(expr)
    var = self.fresh_var(expr.type, name)
    self.assign(var, expr)
    return var

  def int(self, x):
    return const_int(x)

  def bool(self, x):
    return const_bool(x)

  def zero(self, t = Int32, name = "counter"):
    return self.assign_temp(zero(t), name)

  def zero_i32(self, name = "counter"):
    return self.zero(t = Int32, name = name)

  def zero_i64(self, name = "counter"):
    return self.zero(t = Int64, name = name)

  def cast(self, expr, t):
    if expr.type == t:
      return expr
    assert isinstance(t, ScalarT), \
        "Can't cast %s : %s to non-scalar type %s" % (expr, expr.type, t)
    return self.assign_temp(Cast(expr, type = t), "cast_%s" % t)

  def index(self, arr, idx, temp = True, name = None):
    """Index into array or tuple differently depending on the type"""
    
    temp = temp or name is not None
    
    arr_t = arr.type

    if isinstance(arr_t, ScalarT):
      # even though it's not correct externally, it's
      # often more convenient to treat indexing
      # into scalars as the identity function.
      # Just be sure to catch this as an error in
      # the user's code earlier in the pipeline.
      return arr
    if isinstance(arr_t, TupleT):
      if isinstance(idx, Const):
        idx = idx.value

      assert isinstance(idx, int), \
          "Index into tuple must be an integer, got %s" % idx
      if isinstance(idx, Const):
        idx = idx.value
      proj = self.tuple_proj(arr, idx)
      if temp:
        return self.assign_temp(proj, "tuple_elt%d" % idx if name is None else name)
      else:
        return proj

    if self.is_tuple(idx):
      indices = self.tuple_elts(idx)
    elif hasattr(idx, '__iter__'):
      indices = tuple(map(wrap_if_constant,idx))
    else:
      indices = (wrap_if_constant(idx),)

    n_required = arr_t.rank
    n_indices = len(indices)
    if n_indices < n_required:
      # all unspecified dimensions are considered fully sliced
      extra = (syntax_helpers.slice_none,) * (n_required - n_indices)
      indices = indices + extra

    if len(indices) > 1:
      idx = self.tuple(indices, "index_tuple" if name is None else name)
    else:
      idx = indices[0]

    t = arr_t.index_type(idx.type)
    idx_expr = Index(arr, idx, type=t)
    if temp:
      return self.assign_temp(idx_expr, "array_elt" if name is None else name)
    else:
      return idx_expr

  def index_along_axis(self, arr, axis, idx, name=None):
    if arr.type.__class__ is not ArrayT:
      return arr
    assert isinstance(axis, int), \
        "Axis must be a known constant int, got: " + str(axis)
    indices = []
    for i in xrange(arr.type.rank):
      if i == axis:
        indices.append(syntax_helpers.wrap_if_constant(idx))
      else:
        indices.append(syntax_helpers.slice_none)

    index_tuple = self.tuple(indices, "indices")


    result_t = arr.type.index_type(index_tuple.type)
    idx_expr = Index(arr, index_tuple, type=result_t)
    if name:
      return self.assign_temp(idx_expr, name)
    else:
      return idx_expr

  def setidx(self, arr, idx, v):
    self.assign(self.index(arr, idx, temp=False), v)

  def prim(self, prim_fn, args, name=None):
    args = wrap_constants(args)
    arg_types = get_types(args)
    upcast_types = prim_fn.expected_input_types(arg_types)
    result_type = prim_fn.result_type(upcast_types)
    upcast_args = [self.cast(x, t) for (x,t) in zip(args, upcast_types)]
    prim_call = PrimCall(prim_fn, upcast_args, type=result_type)
    if name:
      return self.assign_temp(prim_call, name)
    else:
      return prim_call

  def pick_first(self, x, y):
    """Return x but first cast it to the common type of both args"""

    return self.cast(x, x.type.combine(y.type))

  def pick_second(self, x, y):
    """Return y but first cast it to the common type of both args"""

    return self.cast(y, x.type.combine(y.type))

  def pick_const(self, x, y, c):
    """Return a constant cast to the common type of both args"""

    return self.cast(syntax_helpers.wrap_if_constant(c), x.type.combine(y.type))

  def add(self, x, y, name = None):
    if syntax_helpers.is_zero(x):
      return self.pick_second(x,y)
    elif syntax_helpers.is_zero(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value + y.value)
    else:
      return self.prim(prims.add, [x,y], name)

  def sub(self, x, y, name = None):
    if syntax_helpers.is_zero(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value - y.value)
    else:
      return self.prim(prims.subtract, [x,y], name)

  def mul(self, x, y, name = None):
    if syntax_helpers.is_one(x):
      return self.pick_second(x,y)
    elif syntax_helpers.is_one(y):
      return self.pick_first(x,y)
    elif syntax_helpers.is_zero(x) or syntax_helpers.is_zero(y):
      return self.pick_const(x, y, 0)
    else:
      return self.prim(prims.multiply, [x,y], name)

  def div(self, x, y, name = None):
    if syntax_helpers.is_one(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value / y.value)
    else:
      return self.prim(prims.divide, [x,y], name)

  def safediv(self, x, y, name = None):
    top = self.add(x, y)
    top = self.sub(top, one(top.type))
    return self.div(top, y, name = name)


  def mod(self, x, y, name = None):
    if syntax_helpers.is_one(y):
      return self.pick_const(x, y, 0)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value % y.value)
    else:
      return self.prim(prims.mod, [x,y], name)

  def lt(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(False)
    else:
      return self.prim(prims.less, [x,y], name)

  def lte(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(True)
    else:
      return self.prim(prims.less_equal, [x,y], name)

  def gt(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(False)
    else:
      return self.prim(prims.greater, [x,y], name)

  def gte(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(True)
    else:
      return self.prim(prims.greater_equal, [x,y], name)

  def eq(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(True)
    else:
      return self.prim(prims.equal, [x,y], name)

  def neq(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return syntax_helpers.const_bool(False)
    return self.prim(prims.not_equal, [x,y], name)

  def min(self, x, y, name = None):
    assert x.type == y.type, \
        "Type mismatch between %s and %s" % (x, y)
    if x.__class__ is Const and y.__class__ is Const:
      return x if x.value < y.value else y 
    
    if name is None:
      name = "min_temp"
    result = self.fresh_var(x.type, name)
    cond = self.lte(x, y)
    merge = {result.name:(x,y)}
    self.blocks += syntax.If(cond, [], [], merge)
    return result 

  def max(self, x, y, name = None):
    assert x.type == y.type, \
        "Type mismatch between %s and %s" % (x, y)
    if x.__class__ is Const and y.__class__ is Const:
      return x if x.value > y.value else y 
    if name is None:
      name = "min_temp"
    result = self.fresh_var(x.type, name)
    cond = self.gte(x, y)
    merge = {result.name:(x,y)}
    self.blocks += syntax.If(cond, [], [], merge)
    return result 

  def attr(self, obj, field, name = None):
    if name is None:
      name = field
    obj_t = obj.type
    if obj.__class__ is Struct:
      pos = obj_t.field_pos(name)
      result =  obj.args[pos]
    elif obj.__class__ is ArrayView:
      return getattr(obj, field)
    else:
      assert isinstance(obj_t, StructT), \
        "Can't get attribute '%s' from type %s" % (field, obj_t)
      field_t = obj.type.field_type(field)
      result = Attribute(obj, field, type = field_t)
    if name:
      return self.assign_temp(result, name)
    else:
      return result

  def is_none(self, x):
    return x is None or (isinstance(x, syntax.Expr) and x.type.__class__ is NoneT)

  def is_array(self, x):
    return x.type.__class__ is  ArrayT

  def elt_type(self, x):
    if isinstance(x, Type):
      try:
        return x.elt_type
      except:
        return x
    elif self.is_array(x):
      return x.type.elt_type
    else:
      return x.type

  def ravel(self, x):
    assert self.is_array(x)
    if x.type.rank == 1:
      return x
    print "[Warning] Ravel will fail for discontiguous arrays"
    nelts = self.nelts(x)
    shape = self.tuple((nelts,), 'shape')
    strides = self.tuple((self.int(1),), "strides")
    data = self.attr(x, 'data', 'data_ptr')
    offset = self.attr(x, 'offset')
    t = array_type.make_array_type(x.type.elt_type, 1)
    return syntax.ArrayView(data, shape, strides, offset, nelts, type = t)
  
  def shape(self, array, dim = None):
    if isinstance(array.type, ArrayT):
      shape = self.attr(array, "shape")
      if dim is None:
        return shape
      else:
        dim_t = shape.type.elt_types[dim]
        dim_value = TupleProj(shape, dim, type = dim_t)
        return self.assign_temp(dim_value, "dim%d" % dim)
    else:
      return self.tuple([])

  def strides(self, array, dim = None):
    assert array.type.__class__ is ArrayT
    strides = self.attr(array, "strides")
    if dim is None:
      return strides
    else:
      elt_t = strides.type.elt_types[dim]
      elt_value = TupleProj(strides, dim, type = elt_t)
      return self.assign_temp(elt_value, "stride%d" % dim)

  def tuple(self, elts, name = "tuple", explicit_struct = False):
    if not isinstance(elts, (list, tuple)):
      elts = [elts]
    tuple_t = make_tuple_type(get_types(elts))
    if explicit_struct:
      tuple_expr = Struct(elts, type = tuple_t)
    else:
      tuple_expr = Tuple(elts, type = tuple_t)
    if name:
      result_var = self.assign_temp(tuple_expr, name)
      # cache the simple elements so we can look them up directly
      for (i, elt) in enumerate(elts):
        if self.is_simple(elt):
          self.tuple_elt_cache[(result_var.name, i)] = elt
      return result_var
    else:
      return tuple_expr

  def is_tuple(self, x):
    try:
      return x.type.__class__ is TupleT
    except:
      return False

  def concat_tuples(self, x, y, name = "concat_tuple"):
    if self.is_tuple(x):
      x_elts = self.tuple_elts(x)
    else:
      x_elts = (x,)
    if self.is_tuple(y):
      y_elts = self.tuple_elts(y)
    else:
      y_elts = (y,)

    elts = []
    elts.extend(x_elts)
    elts.extend(y_elts)
    return self.tuple(elts, name = name)

  def tuple_proj(self, tup, idx, explicit_struct = False):
    assert isinstance(idx, (int, long))
    if isinstance(tup, Tuple):
      return tup.elts[idx]
    elif isinstance(tup, tuple):
      return tup[idx]
    elif tup.__class__ is Var and (tup.name, idx) in self.tuple_elt_cache:
      return self.tuple_elt_cache[(tup.name, idx)]
    elif explicit_struct:
      return Attribute(tuple, "elt%d" % idx, type = tup.type.elt_types[idx])
    else:
      return TupleProj(tup, idx, type = tup.type.elt_types[idx])

  def tuple_elts(self, tup, explicit_struct = False):
    nelts = len(tup.type.elt_types)
    return tuple([self.tuple_proj(tup, i, explicit_struct = explicit_struct)
                  for i in xrange(nelts)])

  def closure_elt(self, clos, idx):
    assert isinstance(idx, (int, long))

    if isinstance(clos, Closure):
      return clos.args[idx]
    else:
      return ClosureElt(clos, idx, type = clos.type.arg_types[idx])

  def closure_elts(self, clos):
    if clos.__class__ is TypedFn:
      return ()
    return tuple(self.closure_elt(clos, i)
                 for i in xrange(len(clos.type.arg_types)))

  def get_fn(self, maybe_clos):
    if maybe_clos.__class__ is Closure:
      return maybe_clos.fn
    elif maybe_clos.type.__class__ is ClosureT:
      return maybe_clos.type.fn
    else:
      return maybe_clos

  def closure(self, maybe_fn, extra_args, name = None):
    fn = self.get_fn(maybe_fn)
    old_closure_elts = self.closure_elts(maybe_fn)
    closure_elts = old_closure_elts + tuple(extra_args)
    if len(closure_elts) == 0:
      return fn 
    closure_elt_types = [elt.type for elt in closure_elts]
    closure_t = make_closure_type(fn, closure_elt_types)
    result = Closure(fn, closure_elts, type = closure_t)
    if name:
      return self.assign_temp(result, name)
    else:
      return result

  def prod(self, elts, name = None):
    if self.is_tuple(elts):
      elts = self.tuple_elts(elts)
    if len(elts) == 0:
      return one_i64
    else:
      result = elts[0]
      for e in elts[1:]:
        result = self.mul(result, e, name = name)
      return result

  def alloc_array(self, elt_t, dims, name = "array", explicit_struct = False):
    """
    Given an element type and sequence of expressions denoting each dimension
    size, generate code to allocate an array and its shape/strides metadata. For
    now I'm assuming that all arrays are in row-major, eventually we should make
    the layout an option.
    """

    if self.is_tuple(dims):
      shape = dims
      dims = self.tuple_elts(shape)
    else:
      if not isinstance(dims, (list, tuple)):
        dims = [dims]
      shape = self.tuple(dims, "shape", explicit_struct = explicit_struct)
    rank = len(dims)
    array_t = array_type.make_array_type(elt_t, rank)
    if explicit_struct:
      nelts = self.prod(dims, name = "nelts")
      ptr_t = core_types.ptr_type(elt_t)

      ptr_var = self.assign_temp(Alloc(elt_t, nelts, type = ptr_t), "data_ptr")
      stride_elts = [syntax_helpers.const(1)]

      # assume row-major for now!
      for d in reversed(dims[1:]):
        next_stride = self.mul(stride_elts[0], d, "dim")
        stride_elts = [next_stride] + stride_elts
      strides = self.tuple(stride_elts, "strides", explicit_struct = True)
      array = Struct([ptr_var, shape, strides, zero_i64, nelts], type = array_t)
    else:
      array = AllocArray(shape, type = array_t)
    return self.assign_temp(array, name)

  def return_type(self, fn):
    if isinstance(fn, TypedFn):
      return fn.return_type
    else:
      assert isinstance(fn.type, closure_type.ClosureT), \
          "Unexpected fn type: %s" % fn.type
      assert isinstance(fn.type.fn, TypedFn), \
          "Unexpected fn: %s" % fn.type.fn
      return fn.type.fn.return_type

  def invoke_type(self, closure, args):
    import type_inference
    closure_t = closure.type
    arg_types = syntax_helpers.get_types(args)
    assert all(isinstance(t, core_types.Type) for t in arg_types), \
        "Invalid types: %s" % (arg_types, )
    return type_inference.invoke_result_type(closure_t, arg_types)

  def invoke(self, fn, args):
    import type_inference
    if fn.__class__ is syntax.TypedFn:
      closure_args = []
    else:
      assert isinstance(fn.type, closure_type.ClosureT), \
          "Unexpected function %s with type: %s" % (fn, fn.type)
      closure_args = self.closure_elts(fn)
      arg_types = syntax_helpers.get_types(args)
      fn = type_inference.specialize(fn.type, arg_types)

    import pipeline
    lowered_fn = pipeline.loopify(fn)
    combined_args = tuple(closure_args) + tuple(args)
    call = syntax.Call(lowered_fn, combined_args, type = lowered_fn.return_type)

    return self.assign_temp(call, "call_result")

  def size_along_axis(self, value, axis):
    return self.shape(value, axis)

  def check_equal_sizes(self, sizes):
    pass

  none = syntax_helpers.none
  null_slice = syntax_helpers.slice_none

  # TODO: get rid of that leading underscore to enable this function once
  # shape inference works for all the weird and wacky constructs in our
  # syntax zoo
  def _create_output_array(self, fn, args, extra_dims, name = "output"):
    """
    Given a function and its argument, use shape inference to figure out the
    result shape of the array and preallocate it.  If the result should be a
    scalar, just return a scalar variable.
    """
    try:
      inner_shape_tuple = self.call_shape(fn, args)
    except:
      print "Shape inference failed when calling %s with %s" % (fn, args)
      import sys
      print "Error %s ==> %s" % (sys.exc_info()[:2])
      raise

    if self.is_tuple(extra_dims):
      outer_shape_tuple = extra_dims
    elif isinstance(extra_dims, (list, tuple)):
      outer_shape_tuple = self.tuple(extra_dims)
    else:
      outer_shape_tuple = self.tuple((extra_dims,) if extra_dims else ())

    shape = self.concat_tuples(outer_shape_tuple, inner_shape_tuple)
    elt_t = self.elt_type(self.return_type(fn))
    if len(shape.type.elt_types) > 0:
      return self.alloc_array(elt_t, shape, name)
    else:
      return self.fresh_var(elt_t, name)

  def rank(self, value):
    if self.is_array(value):
      return value.type.rank
    else:
      return 0

  def slice_value(self, start, stop, step):
    slice_t = array_type.make_slice_type(start.type, stop.type, step.type)
    return Slice(start, stop, step, type = slice_t)

  def loop_counter(self, name = "i", start_val = syntax_helpers.zero_i64):
    """
    Generate three SSA variables to use as the before/during/after values
    of a loop counter throughout some loop.

    By default initialize the counter to zero, but optionally start at different
    values using the 'start_val' keyword.
    """

    start_val = syntax_helpers.wrap_if_constant(start_val)
    counter_type = start_val.type

    counter_before = self.assign_temp(start_val, name + "_before")
    counter = self.fresh_var(counter_type, name)
    counter_after = self.fresh_var(counter_type, name + "_after")
    merge = {counter.name:(counter_before, counter_after)}
    return counter, counter_after, merge

  def loop(self, start, niters, loop_body,
            return_stmt = False,
            while_loop = False,
            step = None):
    if step is None:
      step = one(start.type)
      
    if while_loop:
      i, i_after, merge = self.loop_counter("i", start)
      cond = self.lt(i, niters)
      self.blocks.push()
      loop_body(i)
      self.assign(i_after, self.add(i, step))
      body = self.blocks.pop()
      loop_stmt = While(cond, body, merge)
    else:
      var = self.fresh_var(start.type, "i")
      self.blocks.push()
      loop_body(var)
      body = self.blocks.pop()
      loop_stmt = ForLoop(var, start, niters, step, body, {})

    if return_stmt:
      return loop_stmt
    else:
      self.blocks += loop_stmt

  def call_shape(self, maybe_clos, args):
    fn = self.get_fn(maybe_clos)
    closure_args = self.closure_elts(maybe_clos)
    combined_args = tuple(closure_args) + tuple(args)

    if isinstance(fn, Fn):
      # if we're given an untyped function, first specialize it
      import type_inference
      fn = type_inference.specialize(fn, get_types(combined_args))
    abstract_shape = shape_inference.call_shape_expr(fn)
    return shape_codegen.make_shape_expr(self, abstract_shape, combined_args)

  class Accumulator:
    def __init__(self, acc_type, fresh_var, assign):
      self.acc_type = acc_type
      self.fresh_var = fresh_var
      self.assign = assign
      self.start_var = fresh_var(acc_type, "acc")
      self.curr_var = self.start_var

    def get(self):
      return self.curr_var

    def update(self, new_value):
      new_var = self.fresh_var(self.acc_type, "acc")
      self.assign(new_var, new_value)
      self.curr_var = new_var

  def mk_acc(self, init):
    return self.Accumulator(init.type, self.fresh_var, self.assign)
    
  def accumulate_loop(self, start, stop, loop_body, init, return_stmt = False):
    acc = self.mk_acc(init)
    def loop_body_with_acc(i):
      loop_body(acc, i)
    loop_stmt = self.loop(start, stop, loop_body_with_acc, return_stmt = True)
    loop_stmt.merge[acc.start_var.name] = (init, acc.curr_var)
    if return_stmt:
      return loop_stmt, acc.start_var
    else:
      self.blocks += loop_stmt
      return acc.start_var

  def nelts(self, array):
    shape_elts = self.tuple_elts(self.shape(array))
    return self.prod(shape_elts, name = "nelts")

  def array_copy(self, src, dest, return_stmt = False):
    assert self.is_array(dest)
    # nelts = self.nelts(dest)
    shape = self.shape(dest)
    dims = self.tuple_elts(shape)
    rank = len(dims)
    index_vars = []
    def create_loops():
      i = len(index_vars)
      def loop_body(index_var):
        index_vars.append(index_var)
        if i+1 == rank:
          index_tuple = self.tuple(index_vars, "idx")
          lhs = self.index(dest, index_tuple, temp=False)
          rhs = self.index(src, index_tuple, temp=True)
          self.assign(lhs, rhs)
        else:
          self.blocks += create_loops()
      start = syntax_helpers.zero_i64
      stop = dims[i]
      if i > 0 or return_stmt:
        return self.loop(start, stop, loop_body, True)
      else:
        return self.loop(start, stop, loop_body, return_stmt)

    return create_loops()



import syntax_helpers

from array_type import SliceT, make_array_type, ArrayT
from core_types import NoneT, ScalarT, Int64, PtrT, IntT
from syntax import Index, Tuple, Var, ArrayView, Assign, Slice, Struct
from transform import Transform

class LowerIndexing(Transform):
  def pre_apply(self, fn):
    self.bindings = {}

  def tuple_proj(self, tup, idx, explicit_struct = False):
    if tup.__class__ is Var and tup.name in self.bindings:
      stored = self.bindings[tup.name]
      if stored.__class__ is Tuple:
        return stored.elts[idx]
      else:
        return stored.args[idx]
    else:
      return Transform.tuple_proj(self, tup, idx,
                                  explicit_struct = explicit_struct)

  def attr(self, obj, field):
    if obj.__class__ is Var and obj.name in self.bindings:
      stored = self.bindings[obj.name]
      stored_class = stored.__class__
      if stored_class is Struct:
        pos = stored.type.field_pos(field)
        return stored.args[pos]
      elif stored_class  is Slice or stored_class is ArrayView:
        return getattr(stored, field)

    return Transform.attr(self, obj, field)

  def transform_AllocArray(self, expr):
    alloc = self.alloc_array(elt_t = expr.type.elt_type,
                            dims = self.transform_expr(expr.shape),
                            name = "array",
                            explicit_struct = True)
    # recursively transform to turn shape and strides tuples into structs
    return self.transform_expr(alloc)

  def array_slice(self, arr, indices):
    data_ptr = self.attr(arr, "data")
    shape = self.attr(arr, "shape")
    strides = self.attr(arr, "strides")
    elt_offset = self.attr(arr, "offset")
    size = self.attr(arr, "size")

    new_strides = []
    new_shape = []

    for (i, idx) in enumerate(indices):
      stride_i = self.tuple_proj(strides, i)
      shape_i = self.tuple_proj(shape, i)
      idx_t = idx.type
      if isinstance(idx_t, ScalarT):
        offset_i = self.mul(idx, stride_i, "offset_%d" % i)
        elt_offset = self.add(elt_offset, offset_i)
      elif isinstance(idx_t, NoneT):
        new_strides.append(stride_i)
        new_shape.append(shape_i)
      elif isinstance(idx_t, SliceT):
        start = self.attr(idx, "start")
        if isinstance(start.type, NoneT):
          start = syntax_helpers.zero_i64
        stop = self.attr(idx, "stop")
        if isinstance(stop.type, NoneT):
          stop = shape_i
        step = self.attr(idx, "step")
        if isinstance(step.type, NoneT):
          step = syntax_helpers.one_i64

        offset_i = self.mul(start, stride_i, "offset_%d" % i)
        elt_offset = self.add(elt_offset, offset_i)
        dim_i = self.sub(stop, start, "dim_%d" % i)
        # don't forget to cast shape elements to int64
        new_shape.append(self.cast(dim_i, Int64))
        new_strides.append(self.mul(stride_i, step))
      else:
        raise RuntimeError("Unsupported index type: %s" % idx_t)

    elt_t = arr.type.elt_type
    new_rank = len(new_strides)
    new_array_t = make_array_type(elt_t, new_rank)
    new_strides = self.tuple(new_strides, "strides")
    new_shape = self.tuple(new_shape, "shape")
    return ArrayView(data_ptr, new_shape, new_strides,
                            elt_offset, size,
                            type = new_array_t)

  def transform_Index(self, expr):
    arr = self.transform_expr(expr.value)
    idx = self.transform_expr(expr.index)
    idx = self.assign_temp(idx, "idx")

    arr_t = arr.type
    if isinstance(arr_t, PtrT):
      assert isinstance(idx.type, IntT)
      return expr

    assert isinstance(arr_t,  ArrayT), \
        "Unexpected array %s : %s" % (arr, arr.type)

    if self.is_tuple(idx):
      indices = self.tuple_elts(idx)
    else:
      indices = [idx]
    # right pad the index expression with None for each missing index
    n_given = len(indices)
    n_required = arr_t.rank
    if n_given < n_required:
      extra_indices = [syntax_helpers.slice_none] * (n_required - n_given)
      indices.extend(extra_indices)

    # fast-path for the common case when we're indexing
    # by all scalars to retrieve a scalar result
    if syntax_helpers.all_scalars(indices):
      data_ptr = self.attr(arr, "data")
      strides = self.attr(arr, "strides")
      offset_elts = self.attr(arr, "offset")
      for (i, idx_i) in enumerate(indices):
        stride_i = self.tuple_proj(strides, i)
        elts_i = self.mul(stride_i, idx_i, "offset_elts_%d" % i)
        offset_elts = self.add(offset_elts, elts_i, "total_offset")
      return self.index(data_ptr, offset_elts, temp = False)
    else:
      return self.array_slice(arr, indices)

  def transform_Assign(self, stmt):
    lhs = stmt.lhs
    lhs_class = lhs.__class__
    rhs = self.transform_expr(stmt.rhs)
    if lhs_class is Tuple:
      for (i, _) in enumerate(lhs.type.elt_types):
        lhs_i = self.tuple_proj(lhs, i)
        rhs_i = self.tuple_proj(rhs, i)
        # TODO: make this recursive, otherwise nested
        # complex assignments won't get implemented
        assert lhs_i not in (ArrayView, Tuple)
        self.assign(lhs_i, rhs_i)
      return None

    elif lhs_class is Index:
      lhs = self.transform_Index(lhs)
      if isinstance(lhs, ArrayView):
        copy_loop = self.array_copy(src = rhs, dest = lhs, return_stmt = True)
        copy_loop = self.transform_stmt(copy_loop)
        return copy_loop
    elif lhs_class is Var and \
         stmt.rhs.__class__ in (Slice, Struct, ArrayView, Tuple):
      self.bindings[lhs.name] = rhs

    return Assign(lhs, rhs)



import numpy as np

import type_conv

from array_type import ArrayT
from core_types import StructT
from dead_code_elim import DCE
from node import Node
from pipeline_phase import Phase
from simplify import Simplify
from syntax import Var, Fn, TypedFn 
from syntax_helpers import const_int
from syntax_visitor import SyntaxVisitor
from transform import Transform
from tuple_type import TupleT

class AbstractValue(object):
  def __repr__(self):
    return str(self)
  
  def __hash__(self):
    return hash(str(self))

class Unknown(AbstractValue):
  def __str__(self):
    return "unknown"

unknown = Unknown()

class Tuple(AbstractValue):
  def __init__(self, elts):
    self.elts = tuple(elts)
  
  def __str__(self):
    return "Tuple(%s)" % ", ".join(str(elt) for elt in self.elts)
  
  def __eq__(self, other):
    return other.__class__ is Tuple and \
      len(self.elts) == len(other.elts) and \
      all(e1 == e2 for (e1,e2) in zip(self.elts, other.elts))
  
class Array(AbstractValue):
  # mark known strides with integer constants 
  # and all others as unknown
  def __init__(self, strides):
    self.strides = strides
  
  def __str__(self):
    return "Array(strides = %s)" % self.strides
  
  def __eq__(self, other):
    return other.__class__ is Array and \
      self.strides == other.strides

class Struct(AbstractValue):
  def __init__(self, fields):
    self.fields = fields 
    
  def __str__(self):
    return "Struct(%s)" % self.fields
  
  def __eq__(self, other):
    if other.__class__ != Struct:
      return False
    my_fields = set(self.fields.keys())
    other_fields = set(other.fields.keys())
    if my_fields != other_fields:
      return False 
    for f in my_fields:
      if self.fields[f] != other.fields[f]:
        return False
    return True
  
  def __hash__(self):
    return hash(tuple(self.fields.items()))
      
class Const(AbstractValue):
  def __init__(self, value):
    assert isinstance(value, int)
    self.value = value
  
  def __str__(self):
    return str(self.value)
  
  def __eq__(self, other):
    return other.__class__ is Const and self.value == other.value

zero = Const(0)
one = Const(1)

def specialization_const(x, specialize_all = False):
  if x == 0:
    return zero 
  elif x == 1:
    return one
  elif specialize_all:
    return Const(x)
  else:
    return unknown

def abstract_tuple(elts):
  return Tuple(tuple(elts))

def abstract_array(strides):
  return Array(abstract_tuple(strides))

def from_internal_repr(parakeet_type, v):
  if v is None:
    return unknown
  elif hasattr(v, 'contents'):
    v = v.contents
    
  if parakeet_type.__class__ is TupleT:
    elts = []
    for (i,elt_t) in enumerate(parakeet_type.elt_types):
      elt_value = hasattr(v, "elt%d" % i)
      elts.append(from_internal_repr(elt_t, elt_value))
    return abstract_tuple(elts)
  elif parakeet_type.__class__ is ArrayT:
    strides_field = getattr(v, 'strides').contents
    strides = []
    for i in xrange(parakeet_type.rank):
      s = int(getattr(strides_field, 'elt%d'%i))
      strides.append(specialization_const(s))
    return abstract_array(strides)  
  elif isinstance(parakeet_type, StructT):
    fields = {}
    for (field_name, field_type) in parakeet_type._fields_:
      field_value = getattr(v, field_name)
      abstract_field = from_internal_repr(field_type, field_value)
      fields[field_name] = abstract_field
    return Struct(fields)        
  return unknown

def from_python(python_value):
  if isinstance(python_value, np.ndarray):
    elt_size = python_value.dtype.itemsize 
    strides = []
    for s in python_value.strides:
      strides.append(specialization_const(s/elt_size)) 
    return abstract_array(strides)
  elif isinstance(python_value, tuple):
    return abstract_tuple(from_python_list(python_value))
  else:
    parakeet_type = type_conv.typeof(python_value)
    parakeet_value = type_conv.from_python(python_value)
    return from_internal_repr(parakeet_type, parakeet_value)
  
def from_python_list(python_values):
  return [from_python(v) for v in python_values] 

def bind_list(arg_names, abstract_values):
  assert len(arg_names) == len(abstract_values)
  env = {}
  for name, abstract_value in zip(arg_names, abstract_values):
    env[name] = abstract_value
  return env 

class FindConstantStrides(SyntaxVisitor):
  def __init__(self, fn, abstract_values):
    self.env = bind_list(fn.arg_names, abstract_values)
    self.return_value = None 
    
  def visit_generic_expr(self, expr):
    return unknown
  
  def visit_fn(self, fn):
    SyntaxVisitor.visit_fn(self, fn)
    if self.return_value is None:
      return unknown 
    else:
      return self.return_value
  
  def visit_Call(self, expr):
    if expr.fn.__class__ in (Fn, TypedFn): 
      args = self.visit_expr_list(expr.args)
      return  FindConstantStrides(expr.fn, args).visit_fn(expr.fn)
    else:
      return unknown 
      
     
  
  def visit_Var(self, expr):
    return self.env.get(expr.name, unknown)
  
  def visit_Const(self, expr):
    if expr.value == 0:
      return zero
    elif expr.value == 1:
      return one 
    elif isinstance(expr.value, int):
      return Const(expr.value)
    else:
      return unknown
  
  def visit_Attribute(self, expr):
    value = self.visit_expr(expr.value)
    if value.__class__ is Tuple:
      # assume tuples lowered into structs at this point
      pos = expr.value.type.field_pos(expr.name)
      return value.elts[pos]
    elif value.__class__ is Array and expr.name == 'strides':
      return value.strides
    elif value.__class__ is Struct and expr.name in value.fields:
      return value.fields[expr.name]
    else:
      return unknown

  def visit_PrimCall(self, expr):
    abstract_values = self.visit_expr_list(expr.args)
    if all(v.__class__ is Const for v in abstract_values):
      ints = [v.value for v in abstract_values]
      return Const(expr.prim.fn(*ints))
    else:
      return unknown

  def visit_Struct(self, expr):
    if expr.type.__class__ is TupleT:
      return abstract_tuple(self.visit_expr_list(expr.args))
    elif expr.type.__class__ is ArrayT:
      stride_pos = expr.type.field_pos("strides")
      stride_arg = expr.args[stride_pos]
      stride_val = self.visit_expr(stride_arg)
      return Array(stride_val)
    else:
      return unknown   
    
  def visit_Alloc(self, expr):
    return unknown 
  
  def visit_Index(self, expr):
    return unknown 
  
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Var:
      value = self.visit_expr(stmt.rhs)
      assert value is not None, \
        "%s : %s returned None in ConstantStride analysis" % (stmt.rhs, stmt.rhs.node_type()) 
      self.env[stmt.lhs.name] = value 
      
  def visit_Return(self, stmt):
    value = self.visit_expr(stmt.value)
    if self.return_value is None:
      self.return_value = value 
    elif self.return_value != value:
      self.return_value = unknown 
  
class StrideSpecializer(Transform):
  def __init__(self, abstract_inputs):
    Transform.__init__(self)
    self.abstract_inputs = abstract_inputs 
  
  def pre_apply(self, fn):
    analysis = FindConstantStrides(fn, self.abstract_inputs)
    analysis.visit_fn(fn)
    self.env = analysis.env
  
  def transform_Var(self, expr):
    if expr.name in self.env:
      value = self.env[expr.name]
      if value.__class__ is Const:
        return const_int(value.value)
    return expr
  
  def transform_lhs(self, lhs):
    return lhs
  
def has_unit_stride(abstract_value):
  c = abstract_value.__class__
  if c is Array:
    return has_unit_stride(abstract_value.strides)
  elif c is Struct:
    return any(has_unit_stride(field_val) 
               for field_val 
               in abstract_value.fields.itervalues())
  elif c is Tuple:
    return any(has_unit_stride(elt) 
               for elt in abstract_value.elts)
  elif c is Const:
    return abstract_value.value == 1
  else:
    return False
  
_cache = {}
def specialize(fn, python_values, types = None):
  if types is None:
    abstract_values = from_python_list(python_values)
  else:
    # if types are given, assume that the values 
    # are already converted to Parakeet's internal runtime 
    # representation 
    abstract_values = []
    for (t, internal_value) in zip(types, python_values):
      abstract_values.append(from_internal_repr(t, internal_value))
  
  key = (fn.name, tuple(abstract_values))
  if key in _cache:
    return _cache[key]
  elif any(has_unit_stride(v) for v in abstract_values):
    specializer = StrideSpecializer(abstract_values)
    
    transforms = Phase([specializer, Simplify, DCE],
                        memoize = False, copy = True)
    new_fn = transforms.apply(fn)
  else:
    new_fn = fn
  _cache[key] = new_fn
  return new_fn


import core_types
import shape
import shape_from_type
import syntax

from array_type import SliceT, ArrayT
from offset_analysis import OffsetAnalysis
from shape import Var, Const, Shape, Tuple, Closure
from shape import Slice, Scalar, Unknown, Struct
from shape import any_scalar, unknown_value, const, any_value
from shape import combine_list, increase_rank, make_shape
from shape import ConstSlice
from shape_semantics import ShapeSemantics
from syntax_helpers import unwrap_constant
from syntax_visitor import SyntaxVisitor
from tuple_type import TupleT

shape_semantics = ShapeSemantics()
counter = 0
class ShapeInference(SyntaxVisitor):
  def visit_fn(self, fn):
    assert isinstance(fn, syntax.TypedFn)

    self.value_env = {}
    self.equivalence_classes = {}

    self.known_offsets = OffsetAnalysis().visit_fn(fn)

    arg_types = [fn.type_env[name] for name in fn.arg_names]
    input_values = shape_from_type.Converter().from_types(arg_types)
    for n,v in zip(fn.arg_names, input_values):
      self.value_env[n] = v
    self.visit_block(fn.body)

  def unify_scalar_var(self, x, y):
    """
    Unification is different than combining in that it imposes constraints on
    the program. If, for example, we're unifying some scalar that's reached the
    top of the lattice (and thus know nothing about it statically), with a
    scalar known to be some constant-- then the result is we expect both
    variables to be equal to that constant.
    """

    assert isinstance(x, Var), "Expected scalar variable, but got: " + str(x)
    assert isinstance(y, Scalar), "Expected scalar, but got: " + str(y)
    if y == any_scalar:
      return x
    equivs = self.equivalence_classes.get(x, set([]))
    equivs.add(y)
    for var in equivs:
      self.equivalence_classes[var] = equivs
    if isinstance(y, Const):
      for var in equivs:
        self.value_env[var] = y
      return y
    else:
      return var

  def unify_scalar_pairs(self, xs, ys):
    result_elts = []
    for xi, yi in zip(xs.elts, ys.elts):
      result_elts.append(self.unify_scalars(xi, yi))
    return result_elts

  def unify_scalar_list(self, values):
    assert len(values) > 0
    acc = any_scalar
    for v in values:
      acc = self.unify_scalars(acc, v)
    return acc

  def unify_scalars(self, x, y):
    if isinstance(x, Unknown):
      return y
    elif isinstance(y, Unknown):
      return x
    elif isinstance(x, Var):
      return self.unify_scalar_var(x, y)
    elif isinstance(y, Var):
      return self.unify_scalar_var(y, x)
    else:
      raise RuntimeError("Unsupported by unify: %s, %s" % (x,y))

  def visit_merge_loop_start(self, merge):
    for (k, (l, _)) in merge.iteritems():
      self.value_env[k] = self.visit_expr(l)

  def visit_merge(self, merge):
    for (k, (l,r)) in merge.iteritems():
      new_l = self.visit_expr(l)
      new_r = self.visit_expr(r)
      self.value_env[k] = new_l.combine(new_r)

  def visit_expr(self, expr):
    abstract_shape = SyntaxVisitor.visit_expr(self, expr)
    assert abstract_shape is not None, \
        "Unsupported expression in shape inference: %s" % expr.node_type()
    return abstract_shape

  def visit_Alloc(self, expr):
    # alloc doesn't return an array but rather
    # a pointer whose shape properties
    # we don't yet care about here
    return unknown_value

  def visit_Struct(self, expr):
    if isinstance(expr.type, ArrayT):
      shape_tuple = self.visit_expr(expr.args[1])
      return make_shape(shape_tuple.elts)
    elif isinstance(expr.type, TupleT):
      return Tuple(self.visit_expr_list(expr.args))
    elif isinstance(expr.type, SliceT):
      start, stop, step = self.visit_expr_list(expr.args)
      return Slice(start, stop, step)
    else:
      return unknown_value

  def visit_Fn(self, fn):
    return Closure(fn, [])


  def visit_TypedFn(self, fn):
    return Closure(fn, [])

  def shape_from_tuple(self, expr):  
    shape_tuple = self.visit_expr(expr)
    if shape_tuple.__class__ is Tuple:
      return Shape(tuple(shape_tuple.elts))
    else:
      return Shape( (any_scalar,) * expr.type.rank)

  def visit_ArrayView(self, expr):
    return self.shape_from_tuple(expr.shape)
  
  def visit_Reshape(self, expr):
    return self.shape_from_tuple(expr.shape)
  
  def visit_Ravel(self, expr):
    shape = self.visit_expr(expr.array)
    if isinstance(shape, Shape):
      nelts = const(1)
      for dim in shape.dims:
        nelts = shape_semantics.mul(nelts, dim)
      return Shape((nelts,))
    else:
      return any_value 
  
  def visit_Range(self, expr):
    start = self.visit_expr(expr.start)
    stop = self.visit_expr(expr.stop)
    step = self.visit_expr(expr.step)
    slice_value = shape_semantics.slice_value(start, stop, step)
    if slice_value.__class__ is ConstSlice:
      return Shape( (slice_value.nelts,))
    else:
      return Shape( (any_scalar,) )
  
  def visit_ConstArrayLike(self, expr):
    return self.visit_expr(expr.array)
  
  def visit_Slice(self, expr):
    step = self.visit_expr(expr.step)
    if expr.start.__class__ is syntax.Var and \
       expr.stop.__class__ is syntax.Var and \
       step.__class__ is Const:
      start_name = expr.start.name

      stop_name = expr.stop.name
      offsets = self.known_offsets.get(stop_name, [])
      step = step.value if step.value else 1
      for (other_var, offset) in offsets:

        if other_var == start_name:
          nelts = (offset + step - 1) /  step
          # assert False, (start_name, stop_name, offsets)
          return ConstSlice(nelts)

    start = self.visit_expr(expr.start)
    stop = self.visit_expr(expr.stop)
    return shape_semantics.slice_value(start, stop, step)

  def visit_Const(self, expr):
    return Const(expr.value)

  def visit_ClosureElt(self, expr):
    clos = self.visit_expr(expr.closure)
    assert clos.__class__ is Closure, \
        "Unexpected closure shape %s for expression %s" % (clos, expr)
    return clos.args[expr.index]

  def visit_TupleProj(self, expr):
    t = self.visit_expr(expr.tuple)
    assert isinstance(t, Tuple), "Expected tuple type but got %s : %s" % (t, type(t))
    return t.elts[expr.index]

  def visit_Attribute(self, expr):
    v = self.visit_expr(expr.value)
    name = expr.name
    if v.__class__ is Shape and name in ('shape', 'strides'):
      return Tuple(v.dims)
    elif v.__class__ is Tuple and name.startswith('elt'):
      idx = int(name[3:])
      return v[idx]
    elif v.__class__ is Slice:
      return getattr(v, name)
    elif v.__class__ is Struct:
      return v.values[v.fields.index(name)]

    t = expr.value.type.field_type(name)
    if isinstance(t, core_types.ScalarT):
      return any_scalar
    else:
      return any_value

  def visit_PrimCall(self, expr):
    arg_shapes = self.visit_expr_list(expr.args)
    return shape.combine_list(arg_shapes, preserve_const = False)

  def visit_Var(self, expr):
    name = expr.name
    if name in self.value_env:
      return self.value_env[name]
    elif name in self.equivalence_classes:
      for other_name in self.equivalence_classes[name]:
        if other_name in self.value_env:
          return self.value_env[other_name]
    raise RuntimeError("Unknown variable: %s" %  expr)

  def visit_Tuple(self, expr):
    return Tuple(self.visit_expr_list(expr.elts))

  def visit_AllocArray(self, expr):
    shape_tuple = self.visit_expr(expr.shape)
    return make_shape(shape_tuple.elts)

  def visit_Array(self, expr):
    elts = self.visit_expr_list(expr.elts)
    elt = combine_list(elts)
    n = len(elts)
    res = increase_rank(elt, 0, const(n))
    return res

  def visit_Call(self, expr):
    fn = self.visit_expr(expr.fn)
    args = self.visit_expr_list(expr.args)
    return symbolic_call(fn, args)

  def visit_Closure(self, clos):
    assert not isinstance(clos.fn, str), \
        "[ShapeInference] Function names in closures not supported: " + clos.fn
    fn = self.visit_expr(clos.fn)
    closure_arg_shapes = self.visit_expr_list(clos.args)
    if fn.__class__ is Closure:
      closure_arg_shapes = tuple(fn.args) + tuple(closure_arg_shapes)
      fn = fn.fn
    return Closure(fn, closure_arg_shapes)

  def visit_Index(self, expr):
    arr = self.visit_expr(expr.value)
    idx = self.visit_expr(expr.index)
    if arr.__class__ is Tuple and idx.__class__ is Const:
      return arr[idx.value]
    elif arr.__class__ is Shape:
      if isinstance(idx, Scalar):
        return shape.lower_rank(arr, 0)
      elif idx.__class__ is Shape:
        assert len(idx.dims) <= len(arr.dims), \
            "Can't index into rank %d array with rank %d indices" % \
            (len(arr.dims), len(idx.dims))
        dims = [d for d in arr.dims]
        for (i,d) in enumerate(idx.dims):
          dims[i] = d
        return shape.make_shape(dims)
      else:
        return shape_semantics.index(arr, idx)
    assert False, \
        "Can't index (%s) with array shape %s and index shape %s" % \
        (expr, arr, idx)


  def visit_IndexMap(self, expr):
    shape_tuple = self.visit_expr(expr.shape)
    fn = self.visit_expr(expr.fn)
    return shape_semantics.eval_index_map(fn, shape_tuple)
    
  def visit_IndexReduce(self, expr):
    assert False, "IndexReduce not implemented"
    
  def visit_Map(self, expr):
    arg_shapes = self.visit_expr_list(expr.args)
    fn = self.visit_expr(expr.fn)
    axis = unwrap_constant(expr.axis)
    return shape_semantics.eval_map(fn, arg_shapes, axis)
    
  def visit_Reduce(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    arg_shapes = self.visit_expr_list(expr.args)
    init = self.visit_expr(expr.init) if expr.init else None
    axis = unwrap_constant(expr.axis)
    return shape_semantics.eval_reduce(fn, combine, init, arg_shapes, axis)

  def visit_Scan(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    emit = self.visit_expr(expr.emit)
    arg_shapes = self.visit_expr_list(expr.args)
    init = self.visit_expr(expr.init) if expr.init else None
    axis = unwrap_constant(expr.axis)
    return shape_semantics.eval_reduce(fn, combine, emit, init, arg_shapes, axis)

  def visit_AllPairs(self, expr):
    axis = self.unwrap_constant(expr.axis)
    arg_shapes = self.visit_expr_list(expr.args)
    fn = self.visit_expr(expr.fn)
    return shape_semantics.eval_allpairs(fn, arg_shapes, axis)

  def visit_TiledMap(self, expr):
    fn = self.visit_expr(expr.fn)
    args = self.visit_expr_list(expr.args)
    return symbolic_call(fn, args)

  def visit_TiledReduce(self, expr):
    args = self.visit_expr_list(expr.args)
    fn = self.visit_expr(expr.fn)
    return symbolic_call(fn, args)

  def visit_TiledScan(self, expr):
    args = self.visit_expr_list(expr.args)
    fn = self.visit_expr(expr.fn)
    acc_shape = symbolic_call(fn, args)
    emit = self.visit_expr(expr.emit)
    return symbolic_call(emit, [acc_shape])

  def bind(self, lhs, rhs):
    if isinstance(lhs, syntax.Tuple):
      assert isinstance(rhs, Tuple)
      for l,r in zip(lhs.elts, rhs.elts):
        self.bind(l,r)
    elif isinstance(lhs, syntax.Var):
      self.value_env[lhs.name] = rhs

  def visit_Assign(self, stmt):
    rhs = self.visit_expr(stmt.rhs)
    self.bind(stmt.lhs, rhs)

  def visit_Return(self, stmt):
    new_value = self.visit_expr(stmt.value)
    old_value = self.value_env.get("$return", unknown_value)
    combined = old_value.combine(new_value)
    self.value_env["$return"] = combined

  def visit_ForLoop(self, stmt):
    self.value_env[stmt.var.name] = any_scalar
    SyntaxVisitor.visit_ForLoop(self, stmt)

_shape_env_cache = {}
def shape_env(typed_fn):
  key = (typed_fn.name, typed_fn.copied_by)
  if key in _shape_env_cache:
    return _shape_env_cache[key]
  else:
    shape_inference = ShapeInference()
    shape_inference.visit_fn(typed_fn)
    env = shape_inference.value_env
    _shape_env_cache[key] = env
    return env

_shape_cache = {}
def call_shape_expr(typed_fn):
  if isinstance(typed_fn, str):
    typed_fn = syntax.TypedFn.registry[typed_fn]

  key = (typed_fn.name, typed_fn.copied_by)
  if key in _shape_cache:
    return _shape_cache[key]
  else:
    env = shape_env(typed_fn)
    abstract_shape = env.get("$return", Const(None))
    _shape_cache[key] = abstract_shape
    return abstract_shape

def bind(lhs, rhs, env):
  if isinstance(lhs, Var):
    env[lhs] = rhs
  elif isinstance(lhs, Shape):
    assert isinstance(rhs, Shape), "Expected %s, got %s" % (lhs, rhs)
    bind_pairs(lhs.dims, rhs.dims, env)
  elif isinstance(lhs, Closure):
    assert isinstance(rhs, Closure)
    bind_pairs(lhs.args, rhs.args, env)
  elif isinstance(lhs, Tuple):
    assert isinstance(rhs, Tuple)
    bind_pairs(lhs.elts, rhs.elts, env)
  else:
    raise RuntimeError("Unexpected shape LHS: %s" % lhs)

def bind_pairs(xs, ys, env):
  assert len(xs) == len(ys), \
      "Can't bind %s and %s due to unequal lengths" % (xs, ys)
  for (x,y) in zip(xs,ys):
    bind(x,y,env)

def subst(x, env):
  if isinstance(x, Var):
    assert x in env, "Unknown variable %s" % x
    return env[x]
  elif isinstance(x, Scalar):
    return x
  elif isinstance(x, Shape):
    return make_shape(subst_list(x.dims, env))
  elif isinstance(x, Tuple):
    return shape_semantics.tuple(subst_list(x.elts, env))
  elif isinstance(x, Closure):
    return Closure(x.fn, subst_list(x.args, env))
  else:
    raise RuntimeError("Unexpected abstract expression: %s" % x)

def subst_list(xs, env):
  return [subst(x, env) for x in xs]

def symbolic_call(fn, abstract_inputs):
  # result in terms of variables like input0, (shape: input1, input2), etc..
  if fn.__class__ is Closure:
    closure_elts = tuple(fn.args)
    fn = fn.fn
  else:
    closure_elts = ()
  abstract_result_value = call_shape_expr(fn)
  conv = shape_from_type.Converter()
  shape_formals = conv.from_types(fn.input_types)
  
  env = {}
  bind_pairs(shape_formals, closure_elts + tuple(abstract_inputs), env)

  return subst(abstract_result_value, env)



import numpy as np
import parakeet 

from parakeet import jit 
from testing_helpers import run_local_tests, expect

@jit 
def float32_cast(x):
  return parakeet.float32(x)

def test_int_to_float32():
  res = float32_cast(0)
  assert type(res) == np.float32 
  assert res == 0.0
  res = float32_cast(1)
  assert type(res) == np.float32 
  assert res == 1.0
  res = float32_cast(-1)
  assert type(res) == np.float32 
  assert res == -1.0

def test_float_to_float32():
  res = float32_cast(0.0)
  assert type(res) == np.float32 
  assert res == 0.0
  res = float32_cast(1.0)
  assert type(res) == np.float32 
  assert res == 1.0
  res = float32_cast(-1.0)
  assert type(res) == np.float32 
  assert res == -1.0
  

def test_bool_to_float32():
  res = float32_cast(False)
  assert type(res) == np.float32 
  assert res == 0.0
  res = float32_cast(True)
  assert type(res) == np.float32 
  assert res == 1.0

@jit 
def uint8_cast(x):
  return parakeet.uint8(x)

def test_int_to_uint8():
  res = uint8_cast(0)
  assert type(res) == np.uint8 
  assert res == 0.0
  res = uint8_cast(1)
  assert type(res) == np.uint8 
  assert res == 1
  res = uint8_cast(-1)
  assert type(res) == np.uint8 
  assert res == 255

def test_float_to_uint8():
  res = uint8_cast(0.0)
  assert type(res) == np.uint8 
  assert res == 0
  res = uint8_cast(1.0)
  assert type(res) == np.uint8 
  assert res == 1
  res = uint8_cast(-1.0)
  assert type(res) == np.uint8 
  assert res == 255
  

def test_bool_to_uint8():
  res = uint8_cast(False)
  assert type(res) == np.uint8 
  assert res == 0
  res = uint8_cast(True)
  assert type(res) == np.uint8 
  assert res == 1
  
@jit
def type_as_arg(t):
  return t(1)
  
def test_type_as_arg():
  int_res = type_as_arg(np.dtype('int16'))
  assert int_res == 1
  assert type(int_res) == np.int16
  
  float_res = type_as_arg(np.dtype('float64'))
  assert float_res == 1.0
  assert type(float_res) == np.float64
  
@jit
def type_as_default_arg(n, t=parakeet.float64):
  return t(n)
  
def test_type_as_default_arg():
  float_res = type_as_default_arg(-10)
  assert float_res == -10.0
  assert type(float_res) == np.float64
  
  uint_res = type_as_default_arg(10, np.dtype('uint64'))
  assert uint_res == 10
  assert type(uint_res) == np.uint64

@jit
def call_type_conv(n, t):
  return type_as_default_arg(n, t)

def test_call_type_conv():
  float_res = call_type_conv(-1000, parakeet.float64)
  assert type(float_res) == np.float64
  assert float_res == -1000.0, "Expected -1000.0, got %s" % float_res

@jit
def type_as_value(n):
  t2 = np.bool8
  return t2(n)


def test_type_as_value():
  expect(type_as_value, [1], True)
  expect(type_as_value, [0], False)
  
if __name__ == '__main__':
  run_local_tests()



import numpy as np

from testing_helpers import expect, run_local_tests

shape_1d = 40
ints_1d = np.arange(shape_1d)
floats_1d = np.arange(shape_1d, dtype='float')
bools_1d = ints_1d % 2

vecs = [ints_1d, floats_1d, bools_1d]

shape_2d = (4,10)
matrices = [np.reshape(vec, shape_2d) for vec in vecs]

shape_3d = (4,5,2)
tensors = [np.reshape(mat, shape_3d) for mat in matrices]

def index_1d(x, i):
  return x[i]

def test_index_1d():
  for vec in vecs:
    expect(index_1d, [vec, 20], vec[20])

def index_2d(x, i, j):
  return x[i, j]

def test_index_2d():
  for mat in matrices:
    expect(index_2d, [mat, 2, 5], mat[2,5])

def index_3d(x, i, j, k):
  return x[i, j, k]

def test_index_3d():
  for x in tensors:
    expect(index_3d, [x, 2, 2, 1], x[2,2,1])

def set_idx_1d(arr,i,val):
  arr[i] = val
  return arr

def test_set_idx_1d():
  idx = 10
  for vec in vecs:
    vec1, vec2 = vec.copy(), vec.copy()
    val = -vec[idx]
    vec2[idx] = val
    expect(set_idx_1d, [vec1, idx, val], vec2)

def set_idx_2d(arr,i,j,val):
  arr[i, j] = val
  return arr

def test_set_idx_2d():
  i = 2
  j = 2
  for mat in matrices:
    mat1, mat2 = mat.copy(), mat.copy()
    val = -mat[i,j]
    mat2[i,j] = val
    expect(set_idx_2d, [mat1, i, j, val], mat2)

def set_idx_3d(arr, i, j, k, val):
  arr[i, j, k] = val
  return arr

def test_set_idx_3d():
  i = 2
  j = 3
  k = 1
  for x in tensors:
    x1, x2 = x.copy(), x.copy()
    val = -x[i, j, k]
    x2[i, j, k] = val
    expect(set_idx_3d, [x1, i, j, k, val], x2)

if __name__ == '__main__':
  run_local_tests()




import platform 
impl = platform.python_implementation()

from timer import timer 

print "Running under", impl
running_pypy = impl == 'PyPy'

if running_pypy:
  import numpypy as np 
  # crazy that NumPyPy doesn't support this
  def allclose(x,y):
    if x.shape != y.shape:
      return False
    err = ((x-y)**2).mean()
    return err < 0.000001
  def empty_like(x):
    return np.empty(x.shape, dtype=x.dtype)
else:
  import numpy as np 
  allclose = np.allclose
  empty_like = np.empty_like 
import time 



  
k = 7
width, height = 1024,768
if running_pypy:
  from random import randint
  image = np.array([[randint(0,256) for _ in xrange(width)] for _ in xrange(height)])
else:
  image = np.random.randint(0, 256,  (width, height)) / 256.0



def run(fn, name, imshow=False):
  print 
  print "---", name 
  for (prefix, wrapper) in [('parakeet-', jit), ('numba-', autojit)]:
    try:
      wrapped_fn = wrapper(fn)
      with timer(prefix + name + '-compile', True):
        wrapped_fn(image[:1, :1], k)
      with timer(prefix + name, False):
        result = wrapped_fn(image, k)
      if imshow:
        import pylab
        pylab.imshow(image)
        pylab.figure()
        pylab.imshow(result)
        pylab.figure()
        pylab.imshow(scipy_result)
        pylab.show()
      if not running_pypy:
        assert allclose(result, scipy_result)
    except KeyboardInterrupt:
      raise   
    except:
      print "%s failed" % (prefix+name)
      import sys 
      print sys.exc_info()[1]

def dilate_naive(x, k):
  m,n = x.shape
  y = empty_like(x)
  for i in xrange(m):
    for j in xrange(n):
      currmax = x[i,j]
      for ii in xrange(max(0, i-k/2), min(m, i+k/2+1)):
        for jj in xrange(max(0, j-k/2), min(n, j+k/2+1)):
          elt = x[ii,jj]
          if elt > currmax:
            currmax = elt
      y[i,j] = currmax
  return y  

# Numba doesn't yet support min/max so try using inline if expressions
def dilate_naive_inline(x,k):
  m,n = x.shape
  y = empty_like(x)
  for i in xrange(m):
    start_i = i - k/2
    stop_i = i + k/2 + 1
    for j in xrange(n):
      currmax = x[i,j]
      start_j = j - k/2
      stop_j = j + k/2 + 1
      for ii in xrange(start_i if start_i >= 0 else 0, stop_i if stop_i <= m else m):
        for jj in xrange(start_j if start_j >= 0 else 0, stop_j if stop_j <=n else n):
          elt = x[ii,jj]
          if elt > currmax:
            currmax = elt
      y[i,j] = currmax
  return y  

 
def dilate_decompose_loops(x, k):
  m,n = x.shape
  y = empty_like(x)
  for i in xrange(m):
    for j in xrange(n):
      left_idx = max(0, i-k/2)
      right_idx = min(m, i+k/2+1) 
      currmax = x[left_idx, j]
      for ii in xrange(left_idx+1, right_idx):
        elt = x[ii, j]
        if elt > currmax:
          currmax = elt 
      y[i, j] = currmax 
  z = empty_like(x)
  for i in xrange(m):
    for j in xrange(n):
      left_idx = max(0, j-k/2)
      right_idx = min(n, j+k/2+1)
      currmax = y[i,left_idx]  
      for jj in xrange(left_idx+1, right_idx):
        elt = y[i,jj]
        if elt > currmax:
          currmax = elt
      z[i,j] = currmax
  return z 

def dilate_decompose_loops_inline(x, k):
  m,n = x.shape
  y = empty_like(x)
  for i in xrange(m):
    start_i = i-k/2
    stop_i = i+k/2+1 
    for j in xrange(n):
      left_idx = start_i if start_i >= 0 else 0
      right_idx = stop_i if stop_i <= m else m 
      currmax = x[left_idx, j]
      for ii in xrange(left_idx+1, right_idx):
        elt = x[ii, j]
        if elt > currmax:
          currmax = elt 
      y[i, j] = currmax 
  z = empty_like(x)
  for i in xrange(m):
    for j in xrange(n):
      start_j = j-k/2
      stop_j = j+k/2+1
      left_idx = start_j if start_j >= 0 else 0
      right_idx = stop_j if stop_j <= n else n
      currmax = y[i,left_idx]  
      for jj in xrange(left_idx+1, right_idx):
        elt = y[i,jj]
        if elt > currmax:
            currmax = elt
      z[i,j] = currmax
  return z 


def dilate_1d_naive(x_strip,  k):
  """
  Given a 1-dimensional input and 1-dimensional output, 
  fill output with 1d dilation of input 
  """
  nelts = len(x_strip)
  y_strip = empty_like(x_strip)
  half = k / 2 
  for idx in xrange(nelts):
    left_idx = max(idx-half,0)
    right_idx = min(idx+half+1, nelts)
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[idx] = currmax 
  return y_strip

def dilate_decompose(x, k): 
  m,n = x.shape
  y = np.array([dilate_1d_naive(x[row_idx, :], k) for row_idx in xrange(m)])
  return np.array([dilate_1d_naive(y[:, col_idx], k) for col_idx in xrange(n)]).T

def dilate_1d_interior(x_strip, k):
  
  nelts = len(x_strip)
  y_strip = empty_like(x_strip)
  half = k / 2 
  
  interior_start = half+1
  interior_stop = max(nelts-half, interior_start)
  
  # left boundary
  for i in xrange(min(half+1, nelts)):
    left_idx = max(i-half,0)
    right_idx = min(i+half+1, nelts)
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[i] = currmax 
    
  #interior 
  for i in xrange(interior_start, interior_stop):
    left_idx = i-half
    right_idx = i+half+1
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[i] = currmax 
  
  # right boundary
  for i in xrange(interior_stop, nelts):
    left_idx = max(i-half, 0)
    right_idx = nelts
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[i] = currmax 
  return y_strip 

def dilate_decompose_interior(x, k): 
  m,n = x.shape
  y = np.array([dilate_1d_interior(x[row_idx, :],k) for row_idx in xrange(m)])
  return np.array([dilate_1d_interior(y[:, col_idx],k) for col_idx in xrange(n)]).T
 
if __name__ == '__main__':

  if not running_pypy: 
    import scipy.ndimage
    with timer('scipy'):
      scipy_result = scipy.ndimage.grey_dilation(image, k, mode='nearest')
    from numba import autojit 
    from parakeet import jit 
    
    run(dilate_naive, 'naive', imshow=False)
    run(dilate_naive_inline, 'naive-inline')
    run(dilate_decompose_loops, 'decompose-loops')    
    run(dilate_decompose_loops_inline, 'decompose-loops-inline')
    run(dilate_decompose, 'decompose-slices' )
    run(dilate_decompose_interior, 'decompose-interior')

  with timer('cpython-naive'):
    dilate_naive(image, k,)
  with timer('cpython-naive-inline'):
    dilate_naive_inline(image, k)
  with timer('cpython-decompose-loops'):
    dilate_decompose_loops(image, k)
  with timer('cpython-decompose-loops-inline'):
    dilate_decompose_loops_inline(image, k)



import syntax_helpers

from collect_vars import collect_var_names_list
from syntax import Assign, Const, Index, PrimCall, Tuple, TupleProj, Var
from transform import Transform
from use_analysis import use_count

class DCE(Transform):
  def __init__(self):
    Transform.__init__(self, reverse = True)

  def pre_apply(self, fn):
    self.use_counts = use_count(fn)
    
  def is_live(self, name):
    
    count = self.use_counts.get(name)
    return count and count > 0

  def is_live_lhs(self, lhs):
    c = lhs.__class__
    if c is Var:
      return self.is_live(lhs.name)
    elif c is Tuple:
      return any(self.is_live_lhs(elt) for elt in lhs.elts)
    elif isinstance(lhs, (str, tuple)):
      assert False, "Raw data? This ain't the stone age, you know."
    else:
      return True

  def decref(self, expr):
    for var_name in collect_var_names_list(expr):
      self.use_counts[var_name] -= 1

  def transform_merge(self, phi_nodes):
    new_merge = {}
    for (var_name, (l,r)) in phi_nodes.iteritems():

      
      if self.is_live(var_name):

        new_merge[var_name] = (l,r)
      else:
        self.decref(l)
        self.decref(r)

    return new_merge

  def transform_Assign(self, stmt):
    if self.is_live_lhs(stmt.lhs):
      return stmt
    self.decref(stmt.rhs)
    return None

  def is_dead_loop(self, cond, body, merge):
    # pattern match to find loops which only
    # increment a counter without any visible effects
    if syntax_helpers.is_false(cond):
      return True

    rhs_counts = {}
    if cond.__class__ is not Var:
      return False
    rhs_counts[cond.name] = 1
    def process_rhs(expr):
      if expr.__class__ is Var:
        rhs_counts[expr.name] = rhs_counts.get(expr.name, 0) + 1
      elif expr.__class__ is Const:
        pass
      elif expr.__class__ is PrimCall:
        for arg in expr.args:
          process_rhs(arg)
      elif expr.__class__ is Tuple:
        for elt in expr.elts:
          process_rhs(elt)
      elif expr.__class__ is TupleProj:
        process_rhs(expr.tuple)
      elif expr.__class__ is Index:
        process_rhs(expr.value)
        process_rhs(expr.index)
      else:
        return False

    for stmt in body:
      # if statements are anything other than x = safe_expr then
      # there might be externally visible effects to this loop
      if stmt.__class__ is not Assign or stmt.lhs.__class__ is not Var:
        return False
      process_rhs(stmt.rhs)

    for output_name in merge.iterkeys():
      if self.use_counts[output_name] != rhs_counts.get(output_name, 0):
        return False
    return True

  def transform_While(self, stmt):
    # expressions don't get changed by this transform
    new_body = self.transform_block(stmt.body)
    new_merge = self.transform_merge(stmt.merge)
    if self.is_dead_loop(stmt.cond, new_body, new_merge):
      return None
    stmt.body = new_body
    stmt.merge = new_merge
    return stmt

  def transform_If(self, stmt):
    cond = stmt.cond

    # Process the phi-merge first 
    # so that variables dead after the If 
    # statement can become dead inside it 
    stmt.merge = self.transform_merge(stmt.merge)

    stmt.true = self.transform_block(stmt.true)

    stmt.false = self.transform_block(stmt.false)

    if len(stmt.merge) == 0 and len(stmt.true) == 0 and \
        len(stmt.false) == 0:
      return None
    elif syntax_helpers.is_true(cond):
      for name, (_, v) in stmt.merge.iteritems():
        self.assign(Var(name, type = v.type), v)
      self.blocks.extend_current(reversed(stmt.true))
      return None
    elif syntax_helpers.is_false(cond):
      for name, (v, _) in stmt.merge.items():
        self.assign(Var(name, type = v.type), v)
      self.blocks.extend_current(reversed(stmt.false))
      return None
    return stmt

  def transform_Return(self, stmt):
    return stmt
  
  def transform_ExprStmt(self, stmt):
    if self.is_simple(stmt.value):
      return None 
    else:
      return stmt 

  def transform_ForLoop(self, stmt):
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge)
    if len(stmt.body) > 0 or len(stmt.merge) > 0:
      return stmt

  def post_apply(self, fn):
    type_env = {}
    for (name,t) in fn.type_env.iteritems():
      if self.is_live(name):
        type_env[name] = t
    fn.type_env = type_env
    Transform.post_apply(self, fn)
    return fn

def dead_code_elim(fn):
  return DCE().apply(fn)



import numpy as np 
import lib 
import prims
import math  

property_mappings = {
  'dtype' : lib.elt_type,                
  # 'imag' : lib.imag,      
  'itemsize' : lib.itemsize,
  'real' :  lib.identity, # ain't no complex numbers yet 
  'shape' : lib.shape, 
  'size' : lib.size,   
  # 'strides' : lib.strides, 
  'ndim' : lib.rank, 
  'T' : lib.transpose,     
}

method_mappings = {
  'fill' : lib.fill, 
  'any' : lib.any, 
  'all' : lib.all, 
  'argmax' : lib.argmax, 
  # 'argsort' : lib.argsort, 
  'copy' : lib.copy, 
  'cumprod' : lib.cumprod, 
  'cumsum' : lib.cumsum, 
  # 'diagonal' : lib.diagonal, 
  'min' : lib.min, 
  'max' : lib.max, 
  'ravel' : lib.ravel, 
  'flatten' : lib.ravel, 
  'sum' : lib.sum, 
}

function_mappings = {
                     
  map : lib.map, 
  reduce : lib.reduce, 
 
  np.array : lib.identity, 
  tuple : lib._builtin_tuple, 
 
  int : lib.int64, 
  long : lib.int64, 
  np.int8 : lib.int8, 
  np.int16 : lib.int16, 
  np.int32 : lib.int32, 
  np.int64 : lib.int64, 
  
  float : lib.float64, 
  np.float32 : lib.float32, 
  np.float64 : lib.float64, 
  
  bool : lib.bool, 
  np.bool : lib.bool, 
  np.bool8 : lib.bool, 
  np.bool_ : lib.bool, 
  
  np.rank : lib.rank, 
  len : lib.alen, 
  np.alen : lib.alen, 
  np.real : lib.real, 
  # np.imag : lib.imag
  np.size : lib.size, 
  
  min : lib._builtin_min,
  max : lib._builtin_max, 
  np.min : lib.min, 
  np.max : lib.max, 
  np.minimum : prims.minimum, 
  np.maximum : prims.maximum, 
  
  np.argmin : lib.argmin, 
  np.argmax : lib.argmax, 
  
  all : lib.all, 
  np.all : lib.all, 
  any : lib.any, 
  np.any : lib.any, 
  
  sum : lib.sum, 
  np.sum : lib.sum, 
  np.prod : lib.prod, 
  np.mean : lib.mean, 
  
  abs : prims.abs, 
  np.abs : prims.abs, 
  
  range : lib.arange, 
  xrange : lib.arange, 
  np.arange  : lib.arange, 
  
  np.empty_like : lib.empty_like, 
  np.empty : lib.empty, 
  np.zeros_like : lib.zeros_like, 
  np.zeros : lib.zeros, 
  np.ones : lib.ones, 
  np.ones_like : lib.ones_like, 
  
  np.add : prims.add, 
  np.subtract : prims.subtract, 
  np.multiply : prims.multiply, 
  np.divide : prims.divide, 
  
  np.logical_and : prims.logical_and, 
  np.logical_not : prims.logical_not, 
  np.logical_or : prims.logical_or,  
  
  np.cos : prims.cos, 
  math.cos : prims.cos, 
  np.sin : prims.sin, 
  math.sin : prims.sin, 
  np.tan : prims.tan, 
  math.tan : prims.tan, 
  
  math.cosh : prims.cosh, 
  np.cosh : prims.cosh, 
  np.sinh : prims.sinh,           
  math.sinh : prims.sinh, 
  np.tanh : prims.tanh, 
  math.tanh : prims.tanh, 
  
}



import ctypes
# python's function type
from types import FunctionType

import type_conv

 
import prims
from core_types import Type, IncompatibleTypes, StructT, Int64

###########################################
#
#  Closures!
#
###########################################

class ClosureT(StructT):
  _members = ['fn', 'arg_types']

  max_id = 0
  # map each distinct closure_type to an integer id
  id_numbers = {}

  def node_init(self):
    if self.arg_types is None:
      self.arg_types = ()
    elif not hasattr(self.arg_types, '__iter__'):
      self.arg_types = tuple([self.arg_types])
    elif not isinstance(self.arg_types, tuple):
      self.arg_types = tuple(self.arg_types)

    self._fields_ = [('fn_id', Int64)]
    for (i, t) in enumerate(self.arg_types):
      self._fields_.append( ('arg%d' % i, t) )

    self.specializations = {}
    if self in self.id_numbers:
      self.id = self.id_numbers[self]
    else:
      self.id = self.max_id
      self.id_numbers[self] = self.id
      self.max_id += 1

  def __hash__(self):
    return hash( (self.fn,) + self.arg_types)

  def __eq__(self, other):
    if self.fn.__class__ != other.fn.__class__:
      return False
    if isinstance(self.fn, str):
      same_fn = (self.fn == other.fn)
    else:
      # checking equality of functions isn't really defined, 
      # so just check that it's the same location 
      same_fn = self.fn is other.fn
    return same_fn and self.arg_types == other.arg_types

  def __str__(self):
    fn_name = self.fn if isinstance(self.fn, str) else self.fn.name
    return "ClosT(%s, {%s})" % (fn_name, ", ".join(str(t)
                                                   for t in self.arg_types))

  def from_python(self, python_fn):
    import ast_conversion
    untyped_fundef = ast_conversion.translate_function_value(python_fn)
    closure_args = untyped_fundef.python_nonlocals()
    closure_arg_types = map(type_conv.typeof, closure_args)

    closure_t = make_closure_type(untyped_fundef, closure_arg_types)
    closure_id = id_of_closure_type(closure_t)

    def field_value(closure_arg):
      obj = type_conv.from_python(closure_arg)
      parakeet_type = type_conv.typeof(closure_arg)
      if isinstance(parakeet_type, StructT):
        return ctypes.pointer(obj)
      else:
        return obj

    converted_args = [field_value(closure_arg) for closure_arg in closure_args]
    return closure_t.ctypes_repr(closure_id, *converted_args)

  def to_python(self, parakeet_fn):
    raise RuntimeError("TODO: Just return a compiled_fn wrapper")

  def combine(self, other):
    if isinstance(other, ClosureSet):
      return other.combine(self)
    elif isinstance(other, ClosureT):
      if self == other:
        return self
      else:
        return ClosureSet(self, other)
    else:
      raise IncompatibleTypes(self, other)

_closure_type_cache = {}
def make_closure_type(fn, closure_arg_types = []):
  closure_arg_types = tuple(closure_arg_types)
  key = (fn, closure_arg_types)
  if key in _closure_type_cache:
    return _closure_type_cache[key]
  else:
    t = ClosureT(fn, closure_arg_types)
    _closure_type_cache[key] = t
    return t

import config

def print_specializations():
  if config.print_specialized_function_names:
    print
    print "FUNCTION SPECIALIZATIONS"
    count = 0
    for ((untyped,closure_types), clos_t) in sorted(_closure_type_cache.items()):
      specializations = clos_t.specializations.items()
      if len(specializations) > 0:
        name = untyped.name if hasattr(untyped, 'name') else str(untyped)
        print "Closure %s %s" % (name, closure_types)
        for (arg_types, typed_fn) in sorted(specializations):
          print "  -- %s ==> %s" % (arg_types, typed_fn.name)
          count += 1
    print
    print "Total: %d function specializations" % count


import atexit
atexit.register(print_specializations)


"""
Map each (untyped fn id, fixed arg) types to a distinct integer so that the
runtime representation of closures just need to carry this ID
"""

closure_type_to_id = {}
id_to_closure_type = {}
max_id = 0

def id_of_closure_type(closure_t):
  global max_id
  assert isinstance(closure_t, ClosureT), \
      "Expected closure type, got: " + str(closure_t)
  if closure_t in closure_type_to_id:
    return closure_type_to_id[closure_t]
  else:
    num = max_id
    max_id += 1
    closure_type_to_id[closure_t] = num
    return num

def closure_type_from_id(num):
  assert num in id_to_closure_type
  return id_to_closure_type[num]

class ClosureSet(Type):
  """
  If multiple closures meet along control flow paths then join them into a
  closure set. This type should not appear by the time we're generating LLVM
  code.
  """

  _members = ['closures']

  def __init__(self, *closures):
    self.closures = set([])
    for clos_t in closures:
      if isinstance(clos_t, ClosureSet):
        self.closures.update(clos_t.closures)
      else:
        assert isinstance(clos_t, ClosureT)
        self.closures.add(clos_t)

  def combine(self, other):
    if isinstance(other, ClosureSet):
      combined_closures = self.closures.union(other.closures)
      if combined_closures != self.closures:
        return ClosureSet(combined_closures)
      else:
        return self
    else:
      raise IncompatibleTypes(self, other)

  def __eq__(self, other):
    return self.closures == other.closures

  def __iter__(self):
    return iter(self.closures)

  def __len__(self):
    return len(self.closures)



from testing_helpers import expect, run_local_tests

def for_range_loop(start, stop, step):
  x = 0
  for _ in range(start, stop, step):
    x = x + 1
  return x

def test_for_range_loop():
  expect(for_range_loop, [1,10,2], for_range_loop(1,10,2))

def count_loop(init, count):
  x = init
  while x < count:
    x = x + 1
  return x

def test_count_loop():
  expect(count_loop, [0, 300], 300)
  expect(count_loop, [0.0, 400], 400.0)
  expect(count_loop, [0.0, 500.0], 500.0)

def nested_double_count(x):
  total = 0
  i = 0
  while i < x:
    j = 0
    total = total + 1
    while j < x:
      total = total + 1
      j = j + 1
    i = i + 1
  return total

def test_nested_double_count():
  expect(nested_double_count, [10], 110)

def nested_mult(x,y):
  total_count = 0
  i = 0
  while i < x:
    j = 0
    while j < y:
      total_count = total_count + 1
      j = j + 1
    i = i + 1
  return total_count

def test_nested_mult():
  expect(nested_mult, [10, 11], 110)

def conditional_loop(x):
  i = 0
  if x:
    while i < 10:
      i = i + 1
  return i

def test_conditional_loop():
  expect(conditional_loop, [True], 10)
  expect(conditional_loop, [False], 0)

if __name__ == '__main__':
  run_local_tests()



#!/usr/bin/env python

from setuptools import setup, Extension

import sys


setup(
    name="parakeet",
    description="Runtime compiler for flight and amusement.",
    long_description='''
Parakeet
=========

An adorable bird that will make your children laugh and spontaneously combust.

''',
    classifiers=['Development Status :: 3 - Alpha',
                 'Topic :: Software Development :: Libraries',
                 'License :: OSI Approved :: BSD License',
                 'Intended Audience :: Developers',
                 'Programming Language :: Python :: 2.7',
                 ],
    author="Alex Rubinsteyn & Eric Hielscher",
    author_email="alexr@cs.nyu.edu",
    license="BSD",
    version="0.13",
    url="http://github.com/iskandr/parakeet",
    packages=[ 'parakeet', 'tests', 'data' , 'examples'],
    package_dir={ '' : '.' },
    requires=[
      'llvmpy', 
      'numpy', 
      'scipy',
    ])



import ctypes
from llvm.ee import GenericValue

import config
import core_types
import llvm_backend
import llvm_types
import stride_specialization
import syntax
import type_conv
import type_inference

from args import ActualArgs
from llvm_context import global_context
from pipeline import lowering 

def python_to_generic_value(x, t):
  if isinstance(t, core_types.FloatT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.real(llvm_t, x)
  elif isinstance(t, core_types.SignedT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int_signed(llvm_t, x)
  elif isinstance(t, core_types.IntT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int(llvm_t, x)
  elif isinstance(t, core_types.PtrT):
    return GenericValue.pointer(x)
  else:
    ctypes_obj = type_conv.from_python(x)
    return GenericValue.pointer(ctypes.addressof(ctypes_obj))

def ctypes_to_generic_value(cval, t):
  if isinstance(t, core_types.FloatT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.real(llvm_t, cval.value)
  elif isinstance(t, core_types.SignedT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int_signed(llvm_t, cval.value)
  elif isinstance(t, core_types.IntT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int(llvm_t, cval.value)
  elif isinstance(t, core_types.NoneT):
    return GenericValue.int(llvm_types.int64_t, 0)
  elif isinstance(t, core_types.PtrT):
    return GenericValue.pointer(ctypes.addressof(cval.contents))
  else:
    return GenericValue.pointer(ctypes.addressof(cval))

def generic_value_to_python(gv, t):
  if isinstance(t, core_types.SignedT):
    return t.dtype.type(gv.as_int_signed() )
  elif isinstance(t, core_types.IntT):
    return t.dtype.type( gv.as_int() )
  elif isinstance(t, core_types.FloatT):
    llvm_t = llvm_types.ctypes_scalar_to_lltype(t.ctypes_repr)
    return t.dtype.type(gv.as_real(llvm_t))
  elif isinstance(t, core_types.NoneT):
    return None
  else:
    addr = gv.as_pointer()
    struct = t.ctypes_repr.from_address(addr)
    return t.to_python(struct)

class CompiledFn:
  def __init__(self, llvm_fn, parakeet_fn,
               exec_engine = global_context.exec_engine):
    self.llvm_fn = llvm_fn
    self.parakeet_fn = parakeet_fn
    self.exec_engine = exec_engine

  def __call__(self, *args):
    actual_types = tuple(map(type_conv.typeof, args))
    expected_types = self.parakeet_fn.input_types
    assert actual_types == expected_types, \
        "Arg type mismatch, expected %s but got %s" % \
        (expected_types, actual_types)
    # calling conventions are that output must be preallocated by the caller'
    ctypes_inputs = [t.from_python(v) for (v,t) in zip(args, expected_types)]
    gv_inputs = [ctypes_to_generic_value(cv, t) for (cv,t) in
                 zip(ctypes_inputs, expected_types)]

    gv_return = self.exec_engine.run_function(self.llvm_fn, gv_inputs)
    return generic_value_to_python(gv_return, self.parakeet_fn.return_type)

def prepare_args(fn, args, kwargs):
  """
  Fetch the function's nonlocals and return an ActualArgs object of both the arg
  values and their types
  """

  
  if isinstance(fn, syntax.Fn):
    untyped = fn
  else:
    import ast_conversion
    # translate from the Python AST to Parakeet's untyped format
    untyped = ast_conversion.translate_function_value(fn)

  nonlocals = list(untyped.python_nonlocals())
  arg_values = ActualArgs(nonlocals + list(args), kwargs)

  # get types of all inputs
  arg_types = arg_values.transform(type_conv.typeof)
  return untyped, arg_values, arg_types

def specialize_and_compile(fn, args, kwargs = {}):
  """
  Translate, specialize, optimize, and compile the given function for the types
  of the supplies arguments.

  Return the untyped, typed, and compiled representation, along with all the
  arguments needed to actually execute.
  """

  untyped, arg_values, arg_types = prepare_args(fn, args, kwargs)

  # propagate types through function representation and all
  # other functions it calls
  typed = type_inference.specialize(untyped, arg_types)
  lowered = lowering.apply(typed)
  if config.stride_specialization:
    lowered = stride_specialization.specialize(lowered, arg_values)

  # compile to native code
  llvm_fn, parakeet_fn, exec_engine = llvm_backend.compile_fn(lowered)
  compiled_fn_wrapper = CompiledFn(llvm_fn, parakeet_fn, exec_engine)
  return untyped, typed, compiled_fn_wrapper, arg_values

def run(fn, *args, **kwargs):
  """
  Given a python function, run it in Parakeet on the supplied args
  """
  untyped, _, compiled, all_args = specialize_and_compile(fn, args, kwargs)
  linear_args = untyped.args.linearize_without_defaults(all_args)
  return compiled(*linear_args)
